# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Data models for db workspaces."""

import datetime as dt
import enum
import functools
import re
from collections.abc import Callable, Generator, Sequence
from typing import (
    Any,
    Self,
    TYPE_CHECKING,
    TypeAlias,
    TypedDict,
    cast,
    override,
)

from django.conf import settings
from django.contrib.postgres.aggregates import ArrayAgg
from django.contrib.postgres.fields import ArrayField
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db import models
from django.db.models import (
    BigIntegerField,
    Count,
    Exists,
    F,
    Func,
    IntegerField,
    Max,
    OuterRef,
    Q,
    QuerySet,
    Subquery,
    UniqueConstraint,
    Value,
)
from django.db.models.expressions import RawSQL
from django.db.models.functions import Greatest
from django.urls import reverse
from django_cte import CTE, with_cte

from debusine.artifacts.models import CollectionCategory
from debusine.db.models import permissions
from debusine.db.models.files import File
from debusine.db.models.permissions import (
    Allow,
    Predicate,
    PredicateFilter,
    Role,
    enforce,
    permission_check,
    permission_filter,
)
from debusine.db.models.scopes import Scope, ScopeRoles
from debusine.db.permissioncontext import PermissionContext
from debusine.utils.typing_utils import copy_signature_from

if TYPE_CHECKING:
    from django.http import HttpRequest
    from django_stubs_ext import WithAnnotations
    from django_stubs_ext.db.models import TypedModelMeta

    from debusine.db.models import Collection
    from debusine.web.views.ui.workspaces import WorkspaceUI
else:
    TypedModelMeta = object

#: Workspace names reserved for use in toplevel URL path components
RESERVED_WORKSPACE_NAMES = frozenset(
    # TODO: trim after !1274
    (
        "accounts",
        "artifact",
        "task-status",
        "user",
        "workers",
        "work-request",
        "workspaces",
    )
)

#: Regexp matching the structure of workspace names
workspace_name_regex = re.compile(r"^[A-Za-z][A-Za-z0-9+._-]*$")


def is_valid_workspace_name(value: str) -> bool:
    """Check if value is a valid workspace name."""
    if value in RESERVED_WORKSPACE_NAMES:
        return False
    return bool(workspace_name_regex.match(value))


def validate_workspace_name(value: str) -> None:
    """Validate workspace names."""
    if not is_valid_workspace_name(value):
        raise ValidationError(
            "%(value)r is not a valid workspace name", params={"value": value}
        )


def workspace_write_permission_check[M: models.Model](
    msg: str,
    *,
    work_request: Allow = Allow.NEVER,
    anonymous: Allow = Allow.NEVER,
) -> Callable[[Predicate[M]], Predicate[M]]:
    """
    Wrap workspace permission checks with additional checks.

    If the check allows work requests, ensure the workspace matches the work
    request's.

    This is intended for permissions related to state changes.
    """

    def wrap(f: Predicate[M]) -> Predicate[M]:
        f = permission_check(
            msg, work_request=work_request, anonymous=anonymous
        )(f)
        if work_request == Allow.NEVER:
            return f

        @functools.wraps(f)
        def wrapper(self: M, pc: PermissionContext) -> bool:
            if pc.permission_checks_disabled:
                return True

            if not f(self, pc):
                return False

            if pc.work_request is None:
                return True

            return self == pc.work_request.workspace

        setattr(wrapper, "error_template", msg)
        setattr(wrapper, "work_request", work_request)
        setattr(wrapper, "anonymous", anonymous)
        return wrapper

    return wrap


def workspace_write_permission_filter[QS: QuerySet[Workspace, Any]](
    *,
    work_request: Allow = Allow.NEVER,
    anonymous: Allow = Allow.NEVER,
) -> Callable[[PredicateFilter[QS]], PredicateFilter[QS]]:
    """
    Wrap workspace permission filters with additional checks.

    If the filter allows work requests, ensure the workspace matches the work
    request's.

    This is intended for permissions related to state changes.
    """

    def wrap(f: PredicateFilter[QS]) -> PredicateFilter[QS]:
        f = permission_filter(work_request=work_request, anonymous=anonymous)(f)
        if work_request == Allow.NEVER:
            return f

        @functools.wraps(f)
        def wrapper(self: QS, pc: PermissionContext) -> QS:
            if pc.permission_checks_disabled:
                return self

            qs = f(self, pc)
            if pc.work_request is not None:
                qs = qs.filter(pk=pc.work_request.workspace.pk)
            return qs

        setattr(wrapper, "work_request", work_request)
        setattr(wrapper, "anonymous", anonymous)
        return wrapper

    return wrap


class WorkspaceRoleBase(permissions.RoleBase):
    """Workspace role implementation."""

    implied_by_scope_roles: frozenset["ScopeRoles"]
    implied_by_workspace_roles: frozenset["WorkspaceRoles"]
    implied_by_public: bool

    @override
    def _setup(self) -> None:
        """Set up implications for a newly constructed role."""
        implied_by_scope_roles: set[ScopeRoles] = set()
        implied_by_workspace_roles: set[WorkspaceRoles] = {
            cast(WorkspaceRoles, self)
        }
        implied_by_public: bool = False
        for i in self.implied_by:
            match i:
                case ScopeRoles():
                    implied_by_scope_roles |= i.implied_by_scope_roles
                case Role():
                    # Resolve a role passed during class definition into its
                    # enum instance
                    wr = self.__class__(i.value)
                    implied_by_scope_roles |= wr.implied_by_scope_roles
                    implied_by_workspace_roles |= wr.implied_by_workspace_roles
                    implied_by_public = (
                        implied_by_public or wr.implied_by_public
                    )
                case "public":
                    implied_by_public = True
                case _:
                    raise ImproperlyConfigured(
                        "Workspace roles do not support"
                        f" implications by {type(i)}"
                    )
        self.implied_by_scope_roles = frozenset(implied_by_scope_roles)
        self.implied_by_workspace_roles = frozenset(implied_by_workspace_roles)
        self.implied_by_public = implied_by_public

    def q(self, pc: PermissionContext) -> Q:
        """Return a Q expression to select workspaces with this role."""
        if not pc.user.is_authenticated:
            if self.implied_by_public:
                return Q(public=True)
            else:
                return Q(pk__in=())

        q = Q(
            roles__group__in=pc.groups(),
            roles__role__in=self.implied_by_workspace_roles,
        )
        if self.implied_by_scope_roles:
            q |= Q(
                scope__in=Scope.objects.filter(
                    roles__group__in=pc.groups(),
                    roles__role__in=self.implied_by_scope_roles,
                )
            )
        if self.implied_by_public:
            q |= Q(public=True)

        return q

    def implies(self, role: "WorkspaceRoles") -> bool:
        """Check if this role implies the given one."""
        return self in role.implied_by_workspace_roles


class WorkspaceRoles(permissions.Roles, WorkspaceRoleBase, enum.ReprEnum):
    """Available roles for a Scope."""

    OWNER = Role(
        "owner",
        implied_by=[ScopeRoles.OWNER],
        description="Full administrative access",
    )
    CONTRIBUTOR = Role(
        "contributor",
        implied_by=[OWNER],
        description="Contribute using permitted workflow templates",
    )

    VIEWER = Role(
        "viewer",
        implied_by=[CONTRIBUTOR, "public", ScopeRoles.OBSERVER],
        description="Read-only access",
    )


WorkspaceRoles.setup()


class ExpirationTimeDict(TypedDict):
    """Additional fields returned by WorkspaceQuerySet.with_expiration_time."""

    expiration_time: dt.datetime


class WorkspaceQuerySet[A](QuerySet["Workspace", A]):
    """Custom QuerySet for Workspace."""

    def with_expiration_time(
        self,
    ) -> "WorkspaceQuerySet[WithAnnotations[Workspace, ExpirationTimeDict]]":
        """Annotate the queryset with expiration times."""
        return self.annotate(
            expiration_time=F("expiration_delay")
            + Greatest("created_at", Max("workrequest__completed_at"))
        )

    def with_role_annotations(self, pc: PermissionContext) -> Self:
        """
        Annotate the queryset with a bool for each role, with its presence.

        The returned annotations are in the form ``"is_{role}"`` and are True
        if the user has the given role on the annotated resource.

        A further ``roles_for`` annotation marks the stringified public key of
        the user used to compute annotations, or the empty string if the
        annotations are for the anonymous user.
        """
        queryset = self.annotate(
            roles_for=Value(str(pc.user.pk) if pc.user.is_authenticated else "")
        )
        for role in WorkspaceRoles:
            queryset = queryset.annotate(
                **{
                    f"is_{role}": Exists(
                        Workspace.objects.filter(pk=OuterRef("pk")).with_role(
                            pc, role
                        )
                    )
                }
            )
        return queryset

    def with_role(self, pc: PermissionContext, role: WorkspaceRoles) -> Self:
        """Keep only resources where the user has the given role."""
        if pc.user.is_authenticated and pc.user.is_system:
            return self
        return self.filter(
            Exists(self.model.objects.filter(role.q(pc), pk=OuterRef("pk")))
        )

    def in_current_scope(self) -> Self:
        """Filter to workspaces in the current scope."""
        from debusine.db.context import ContextConsistencyError, context

        if context.scope is None:
            raise ContextConsistencyError("scope is not set")
        return self.filter(scope=context.scope)

    @permission_filter(work_request=Allow.PASS, anonymous=Allow.PASS)
    def can_display(self, pc: PermissionContext) -> Self:
        """Keep only Workspaces that can be displayed."""
        # This is the same check done in Workspace.set_current, and if changed
        # they need to be kept in sync
        return self.with_role(pc, Workspace.Roles.VIEWER)

    @workspace_write_permission_filter()
    def can_configure(self, pc: PermissionContext) -> Self:
        """Keep only Workspaces that can be configured."""
        return self.with_role(pc, Workspace.Roles.OWNER)

    @workspace_write_permission_filter(work_request=Allow.PASS)
    def can_create_artifacts(self, pc: PermissionContext) -> Self:
        """Keep only Workspaces where the user can create artifacts."""
        return self.with_role(pc, Workspace.Roles.CONTRIBUTOR)

    @workspace_write_permission_filter()
    def can_create_collections(self, pc: PermissionContext) -> Self:
        """Keep only Workspaces where the user can create collections."""
        return self.with_role(pc, Workspace.Roles.OWNER)

    @workspace_write_permission_filter()
    def can_create_work_requests(self, pc: PermissionContext) -> Self:
        """Workspaces where the user can create work requests."""
        return self.with_role(pc, Workspace.Roles.CONTRIBUTOR)

    @workspace_write_permission_filter(work_request=Allow.PASS)
    def can_create_child_workspace(self, pc: PermissionContext) -> Self:
        """Workspaces from which the user can create a child workspace."""
        return self.with_role(pc, Workspace.Roles.CONTRIBUTOR)

    @workspace_write_permission_filter()
    def can_edit_task_configuration(self, pc: PermissionContext) -> Self:
        """Workspaces in which the user can edit task configuration."""
        return self.with_role(pc, Workspace.Roles.OWNER)

    def get_for_context(self, name: str) -> "Workspace":
        """
        Query a workspace for setting into the current context.

        This defaults to context.scope as the scope, and prefetches roles for
        the current user
        """
        # Prevent import loop
        from debusine.db.context import context

        # Note: user permissions are checked by Workspace.set_current
        scope = context.require_scope()
        pc = context.pc

        # Do not annotate with user roles if using a work request token
        if pc.work_request_token:
            return cast("Workspace", self.get(scope=scope, name=name))

        queryset = self.select_related("scope")
        if pc.user.is_authenticated:
            queryset = cast(
                "WorkspaceQuerySet[A]",
                queryset.annotate(
                    user_roles=ArrayAgg(
                        "roles__role",
                        filter=Q(roles__group__in=pc.groups()),
                        distinct=True,
                        default=Value([]),
                    )
                ),
            )
        workspace = queryset.get(scope=scope, name=name)
        assert isinstance(workspace, Workspace)
        return workspace

    def annotate_with_workflow_stats(self) -> "WorkspaceQuerySet[Any]":
        """
        Annotate with workflow statistics.

        * ``running``: number of running workflows
        * ``needs_input``: number of workflows needing input
        * ``completed``: number of completed workflows
        """
        # Prevent circular import loop
        from debusine.db.models.work_requests import WorkRequest

        # Using subqueries so we can use our custom QuerySet filter methods.
        #
        # See https://stackoverflow.com/a/43771738 for an explanation of how
        # the annotations below are constructed.
        return self.annotate(
            running=Subquery(
                WorkRequest.objects.running()
                .workflows()
                .filter(parent__isnull=True)
                .filter(workspace=OuterRef("pk"))
                .values("workspace")
                .annotate(count=Count("pk"))
                .values("count")
            ),
            needs_input=Subquery(
                WorkRequest.objects.needs_input()
                .workflows()
                .filter(parent__isnull=True)
                .filter(workspace=OuterRef("pk"))
                .values("workspace")
                .annotate(count=Count("pk"))
                .values("count")
            ),
            completed=Subquery(
                WorkRequest.objects.filter(
                    status__in=(
                        WorkRequest.Statuses.COMPLETED,
                        WorkRequest.Statuses.ABORTED,
                    )
                )
                .workflows()
                .filter(parent__isnull=True)
                .filter(workspace=OuterRef("pk"))
                .values("workspace")
                .annotate(count=Count("pk"))
                .values("count")
            ),
        )


class WorkspaceManager(models.Manager["Workspace"]):
    """Manager for Workspace model."""

    def get_roles_model(self) -> type["WorkspaceRole"]:
        """Get the model used for role assignment."""
        return WorkspaceRole

    @override
    def get_queryset(self) -> WorkspaceQuerySet[Any]:
        """Use the custom QuerySet."""
        return WorkspaceQuerySet(self.model, using=self._db)


class DeleteWorkspaces:
    """Delete all workspaces in a queryset."""

    def __init__(self, workspaces: WorkspaceQuerySet["Any"]) -> None:
        """Store workspaces to delete and compute affected objects."""
        # Import here to prevent circular dependencies
        from debusine.db import models as dmodels

        self.workspaces = workspaces

        # Since we use on_delete=PROTECT on most models, there may be
        # elements in the model interdependency graphs that we are not
        # deleting yet.
        #
        # It is difficult to test this without having infrastructure to
        # simulate a fully populated database that is kept up to date as
        # new models get added, so for the moment we limit ourselves to
        # adding to this as the need arises.
        self.artifact_relations = dmodels.ArtifactRelation.objects.filter(
            Q(artifact__workspace__in=workspaces)
            | Q(target__workspace__in=workspaces)
        )
        self.artifacts = dmodels.Artifact.objects.filter(
            workspace__in=workspaces
        )
        self.asset_usages = dmodels.AssetUsage.objects.filter(
            workspace__in=workspaces
        )
        self.assets = dmodels.Asset.objects.filter(workspace__in=workspaces)
        self.collection_items = dmodels.CollectionItem.objects.filter(
            Q(parent_collection__workspace__in=workspaces)
            | Q(artifact__workspace__in=workspaces)
            | Q(collection__workspace__in=workspaces)
        )
        self.collections = dmodels.Collection.objects.filter(
            workspace__in=workspaces
        )
        self.file_in_artifacts = dmodels.FileInArtifact.objects.filter(
            artifact__workspace__in=workspaces
        )
        self.file_uploads = dmodels.FileUpload.objects.filter(
            file_in_artifact__in=self.file_in_artifacts
        )
        # Group role tables use CASCADE on the foreign key to group, so we
        # don't need to track where workspace-scoped groups are used and we can
        # just delete them
        self.groups = dmodels.Group.objects.filter(workspace__in=workspaces)
        self.work_requests = dmodels.WorkRequest.objects.filter(
            workspace__in=workspaces
        )
        self.workflow_templates = dmodels.WorkflowTemplate.objects.filter(
            workspace__in=workspaces
        )
        self.workspace_roles = WorkspaceRole.objects.filter(
            resource__in=workspaces
        )

    def perform_deletions(self) -> None:
        """Delete the workspaces and all their resources."""
        # Delete file uploads individually, since we need to schedule
        # unlinks from the upload directory when the transaction is
        # committed.
        for file_upload in self.file_uploads:
            file_upload.delete()

        self.workflow_templates.delete()
        self.collection_items.delete()
        self.work_requests.delete()
        self.collections.delete()
        self.file_in_artifacts.delete()
        self.artifact_relations.delete()
        self.artifacts.delete()
        self.asset_usages.delete()
        self.assets.delete()
        self.workspace_roles.delete()
        self.groups.delete()
        self.workspaces.delete()


def default_workspace() -> "Workspace":
    """Return the default Workspace."""
    return Workspace.objects.get(
        scope__name=settings.DEBUSINE_DEFAULT_SCOPE,
        name=settings.DEBUSINE_DEFAULT_WORKSPACE,
    )


class Workspace(models.Model):
    """Workspace model."""

    Roles: TypeAlias = WorkspaceRoles

    name = models.CharField(
        max_length=255, validators=[validate_workspace_name]
    )
    public = models.BooleanField(default=False)
    default_expiration_delay = models.DurationField(
        default=dt.timedelta(0),
        help_text="minimal time that a new artifact is kept in the"
        " workspace before being expired",
    )

    inherits = models.ManyToManyField(
        "db.Workspace",
        through="db.WorkspaceChain",
        through_fields=("child", "parent"),
        related_name="inherited_by",
    )

    scope = models.ForeignKey(
        Scope, on_delete=models.PROTECT, related_name="workspaces"
    )

    created_at = models.DateTimeField(
        auto_now_add=True, help_text="time the workspace was created"
    )
    expiration_delay = models.DurationField(
        blank=True,
        null=True,
        help_text=(
            "if set, time since the last task completion time"
            " after which the workspace can be deleted"
        ),
    )

    objects = WorkspaceManager.from_queryset(WorkspaceQuerySet)()

    class Meta(TypedModelMeta):
        base_manager_name = "objects"
        constraints = [
            UniqueConstraint(
                fields=["scope", "name"],
                name="%(app_label)s_%(class)s_unique_scope_name",
            ),
        ]

    @override
    def __str__(self) -> str:
        """Return basic information of Workspace."""
        return f"{self.scope.name}/{self.name}"

    @override
    @copy_signature_from(models.Model.save)
    def save(self, **kwargs: Any) -> None:
        """Wrap save with permission checks."""
        if self._state.adding:
            # Create
            enforce(self.scope.can_create_workspace)
        else:
            # Update
            ...  # TODO: check for update permissions

        return super().save(**kwargs)

    def get_absolute_url(self) -> str:
        """Return an absolute URL to this workspace."""
        from debusine.server.scopes import urlconf_scope

        with urlconf_scope(self.scope.name):
            return reverse("workspaces:detail", kwargs={"wname": self.name})

    def ui(self, request: "HttpRequest") -> "WorkspaceUI":
        """Return a UI helper for this instance."""
        from debusine.web.views.ui.workspaces import WorkspaceUI

        return WorkspaceUI.get(request, self)

    def get_absolute_url_configure(self) -> str:
        """Return an absolute URL to configure this workspace."""
        from debusine.server.scopes import urlconf_scope

        with urlconf_scope(self.scope.name):
            return reverse("workspaces:update", kwargs={"wname": self.name})

    def get_absolute_url_configure_inheritance(self) -> str:
        """Return an absolute URL to configure the inheritance chain."""
        from debusine.server.scopes import urlconf_scope

        with urlconf_scope(self.scope.name):
            return reverse(
                "workspaces:update_inheritance", kwargs={"wname": self.name}
            )

    def set_current(self, *, check_visibility: bool = True) -> None:
        """
        Set this as the current workspace.

        This needs to be called after ``context.set_scope`` and
        ``context.set_user``.

        :param check_visibility: if True, check whether this workspace is
          visible.  Views may only set this to False if they promise to
          enforce a more specific permission instead; this is useful for
          views that accept narrowly-scoped tokens, which might allow access
          to a specific resource in a workspace but not to the whole
          workspace.
        """
        from debusine.db.context import ContextConsistencyError, context

        if (old_workspace := context.workspace) is not None:
            raise ContextConsistencyError(
                f"Workspace was already set to {old_workspace}"
            )

        if (scope := context.scope) is None:
            raise ContextConsistencyError("Cannot set workspace before scope")

        if self.scope != scope:
            raise ContextConsistencyError(
                f"workspace scope {self.scope.name!r}"
                f" does not match current scope {scope.name!r}"
            )

        _workspace_roles: set[str] = set()

        if self.public:
            _workspace_roles.add(Workspace.Roles.VIEWER)

        pc = context.pc
        if not context.user.is_authenticated:
            pass
        elif (user_roles := getattr(self, "user_roles", None)) is not None:
            # Use cached version if available
            _workspace_roles.update(user_roles)
        else:
            _workspace_roles.update(self.get_group_roles(pc))

        workspace_roles = Workspace.Roles.from_iterable(_workspace_roles)

        if check_visibility:
            # Check workspace visibility. This is the same as the can_display
            # predicate, and if changed they need to be kept in sync
            expected = Workspace.Roles.VIEWER
            if (
                pc.work_request_token
                or not expected.implied_by_scope_roles.isdisjoint(
                    context.scope_roles
                )
                or any(a.implies(expected) for a in workspace_roles)
            ):
                context._workspace.set(self)
                context._workspace_roles.set(workspace_roles)
            else:
                raise ContextConsistencyError(
                    f"User {context.user} cannot access workspace {self}"
                )
        else:
            context._workspace.set(self)
            context._workspace_roles.set(workspace_roles)

    @permission_check(
        "{user} cannot display workspace {resource}",
        work_request=Allow.PASS,
        anonymous=Allow.PASS,
    )
    def can_display(self, pc: PermissionContext) -> bool:
        """Check if the workspace can be displayed."""
        return self.has_role(pc, Workspace.Roles.VIEWER)

    @workspace_write_permission_check(
        "{user} cannot configure workspace {resource}",
    )
    def can_configure(self, pc: PermissionContext) -> bool:
        """Check if the workspace can be configured."""
        return self.has_role(pc, Workspace.Roles.OWNER)

    @workspace_write_permission_check(
        "{user} cannot create artifacts in {resource}",
        work_request=Allow.PASS,
    )
    def can_create_artifacts(self, pc: PermissionContext) -> bool:
        """Check if the user can create artifacts."""
        return self.has_role(pc, Workspace.Roles.CONTRIBUTOR)

    @workspace_write_permission_check(
        "{user} cannot create collections in {resource}"
    )
    def can_create_collections(self, pc: PermissionContext) -> bool:
        """Check if the user can create collections."""
        return self.has_role(pc, Workspace.Roles.OWNER)

    @workspace_write_permission_check(
        "{user} cannot create work requests in {resource}",
    )
    def can_create_work_requests(self, pc: PermissionContext) -> bool:
        """Check if the user can create work requests."""
        # TODO: At some point we may disallow users with only
        # contributor permissions from creating work requests
        # directly rather than via workflows, but that depends on
        # some other things, such as exactly how we implement custom
        # workflows and making it easy for users to create their own
        # workspaces.
        return self.has_role(pc, Workspace.Roles.CONTRIBUTOR)

    @workspace_write_permission_check(
        "{user} cannot create a child workspace from {resource}",
        work_request=Allow.PASS,
    )
    def can_create_child_workspace(self, pc: PermissionContext) -> bool:
        """Check if the user can create a child workspace from this."""
        return self.has_role(pc, Workspace.Roles.CONTRIBUTOR)

    @workspace_write_permission_check(
        "{user} cannot edit task configuration in {resource}",
    )
    def can_edit_task_configuration(self, pc: PermissionContext) -> bool:
        """Check if the user can edit task configuration."""
        return self.has_role(pc, Workspace.Roles.OWNER)

    def has_role(self, pc: PermissionContext, role: WorkspaceRoles) -> bool:
        """Check if the user has the given role on this Workspace."""
        from debusine.db.context import context

        # Honor implied_by_public immediately as it's independent from the user
        if role.implied_by_public and self.public:
            return True

        if not pc.user.is_authenticated:
            return False

        if pc.user.is_system:
            return True

        if context.pc_is(pc) and context.workspace == self:
            # Allowed as implied by scope roles the user has
            if not context.scope_roles.isdisjoint(role.implied_by_scope_roles):
                return True

            # Allowed as implied by workspace roles the user has
            if any(r.implies(role) for r in context.workspace_roles):
                return True

            # The context has enough information to know that the user does not
            # have the given role on this workspace
            return False
        return Workspace.objects.with_role(pc, role).filter(pk=self.pk).exists()

    # See https://github.com/typeddjango/django-stubs/issues/1047 for the typing
    def get_group_roles(
        self, pc: PermissionContext
    ) -> QuerySet["WorkspaceRole", str]:
        """Get the roles of the user on this workspace."""
        if not pc.user.is_authenticated:
            return WorkspaceRole.objects.none().values_list("role", flat=True)
        else:
            return (
                self.roles.filter(group__in=pc.groups())
                .values_list("role", flat=True)
                .distinct()
            )

    def get_roles(self, pc: PermissionContext) -> frozenset[WorkspaceRoles]:
        """
        Get the effective roles of the user on this workspace.

        If you invoke this method on each result of a whole queryset, you
        should annotate the queryset using ``with_role_annotations`` to greatly
        reduce the number of queries.
        """
        roles: list[WorkspaceRoles] = []
        if ((roles_for := getattr(self, "roles_for", None)) is not None) and (
            (
                roles_for
                and pc.user
                and pc.user.is_authenticated
                and int(roles_for) == pc.user.pk
            )
            # with_role_annotations uses the empty string to indicate that the
            # annotations are computed for the anonymous user
            or (roles_for == "" and not pc.user.is_authenticated)
        ):
            # Use roles from annotations
            roles += [
                role
                for role in WorkspaceRoles
                if getattr(self, f"is_{role}", False)
            ]
        else:
            # Fall back on querying each role
            roles += [
                role for role in WorkspaceRoles if self.has_role(pc, role)
            ]
        return WorkspaceRoles.from_iterable(roles)

    @property
    def expire_at(self) -> dt.datetime | None:
        """Return computed expiration date."""
        if self.expiration_delay is None:
            return None

        # Try to reuse a value precomputed by
        # WorkspaceQuerySet.with_expiration_time()
        # otherwise, compute it and set it

        if not hasattr(self, "expiration_time"):
            expiration_time = (
                Workspace.objects.with_expiration_time()
                .values_list("expiration_time", flat=True)
                .get(pk=self.pk)
            )
            setattr(self, "expiration_time", expiration_time)

        return cast(
            dt.datetime,
            self.expiration_time,  # type: ignore[attr-defined]
        )

    def file_needs_upload(self, fileobj: File) -> bool:
        """
        Return True if fileobj needs to be uploaded to this workspace.

        Before an artifact can be considered complete, all its files must be
        part of the artifact's workspace.  This requires each file to be in
        one of its stores, and also to be in some artifact in the workspace.
        Otherwise, the file must be uploaded, even if it already exists
        somewhere else in debusine; this prevents users obtaining
        unauthorized access to existing file contents.
        """
        from debusine.db.models import FileInArtifact

        if not self.scope.download_file_stores(fileobj).exists():
            return True

        if not FileInArtifact.objects.filter(
            artifact__workspace=self, file=fileobj, complete=True
        ).exists():
            return True

        return False

    def set_inheritance(self, chain: Sequence["Workspace"]) -> None:
        """Set the inheritance chain for this workspace."""
        # Check for duplicates in the chain before altering the database
        seen: set[int] = set()
        for workspace in chain:
            if workspace == self:
                raise ValueError(
                    "inheritance chain contains the workspace itself"
                )

            if workspace.pk in seen:
                raise ValueError(
                    f"duplicate workspace {workspace.name!r}"
                    " in inheritance chain"
                )
            seen.add(workspace.pk)

        WorkspaceChain.objects.filter(child=self).delete()
        for idx, workspace in enumerate(chain):
            WorkspaceChain.objects.create(
                child=self, parent=workspace, order=idx
            )

    def make_inheritance_chain_cte(
        self, visible: CTE
    ) -> Callable[[CTE], "WorkspaceQuerySet[Workspace]"]:
        """
        Make a recursive CTE factory to follow a workspace's inheritance chain.

        :param visible: A Common Table Expression with the results of
          ``Workspace.objects.can_display(user)``, which will be used by
          this CTE.
        :return: a function that can be passed as the first argument to
          ``CTE.recursive``.  The ``path_order`` field in the ``QuerySet``
          it returns allows ordering the results as for a depth-first
          search.
        """

        def chain_cte(cte: CTE) -> WorkspaceQuerySet[Workspace]:
            path_order_field: ArrayField[int, int] = ArrayField(
                IntegerField(), name="path_order"
            )
            path_field: ArrayField[int, int] = ArrayField(
                BigIntegerField(), name="path"
            )
            path_field_contains = path_field.get_lookup("contains")
            assert path_field_contains is not None
            recursive = (
                visible.join(Workspace, id=visible.col.id)
                .filter(id=self.id)
                .values(
                    chain_child=F("id"),
                    chain_parent=F("id"),
                    path_order=RawSQL(
                        "ARRAY[]::integer[]", (), output_field=path_order_field
                    ),
                    path=Func(
                        F("id"),
                        function="ARRAY",
                        template="%(function)s[%(expressions)s]",
                        output_field=path_field,
                    ),
                )
                .union(
                    visible.join(
                        cte.join(WorkspaceChain, child=cte.col.chain_parent),
                        parent_id=visible.col.id,
                    )
                    # Break inheritance loops.
                    .annotate(
                        is_cycle=path_field_contains(
                            cte.col.path,
                            Func(
                                F("parent_id"),
                                function="ARRAY",
                                template="%(function)s[%(expressions)s]",
                                output_field=path_field,
                            ),
                        )
                    )
                    .filter(is_cycle=False)
                    .values(
                        chain_child=F("child_id"),
                        chain_parent=F("parent_id"),
                        path_order=Func(
                            cte.col.path_order,
                            F("order"),
                            function="array_append",
                            output_field=path_order_field,
                        ),
                        path=Func(
                            cte.col.path,
                            F("parent_id"),
                            function="array_append",
                            output_field=path_field,
                        ),
                    ),
                    all=True,
                )
            )
            assert isinstance(recursive, WorkspaceQuerySet)
            return recursive

        return chain_cte

    def get_collection(
        self, *, pc: PermissionContext, category: str, name: str
    ) -> "Collection":
        """
        Look up a collection by category and name.

        If the collection is not found in this workspace, it follows the
        workspace inheritance chain using a depth-first search.

        :param user: user to use for permission checking
        :param category: collection category
        :param name: collection name
        :raises Collection.DoesNotExist: if the collection was not found
        """
        from debusine.db.models import Collection

        visible = CTE(Workspace.objects.can_display(pc), name="visible")
        chain = CTE.recursive(
            self.make_inheritance_chain_cte(visible), name="chain"
        )
        collection = (
            with_cte(
                visible,
                chain,
                select=chain.join(
                    Collection, workspace=chain.col.chain_parent
                ).filter(category=category, name=name),
            )
            # Select the first matching collection in depth-first search
            # order.  This raises Collection.DoesNotExist if there is none.
            .earliest(chain.col.path_order)
        )
        assert isinstance(collection, Collection)
        return collection

    def get_singleton_collection(
        self, *, pc: PermissionContext, category: str
    ) -> "Collection":
        """
        Lookup a singleton collection by category.

        If the collection is not found in this workspace, it follows the
        workspace inheritance chain using a depth-first search.

        :param user: user to use for permission checking
        :param category: collection category
        :raises Collection.DoesNotExist: if the collection was not found
        """
        return self.get_collection(pc=pc, category=category, name="_")

    def list_accessible_collections(
        self,
        pc: PermissionContext,
        category: CollectionCategory,
        *,
        ws_seen: tuple[int, ...] = (),
        coll_seen: set[int] | None = None,
    ) -> Generator["Collection"]:
        """
        List all collections of a given category accessible from a workspace.

        It recurses up the workspace inheritance chain, breaking loops.

        :param user: user to use for accessibility checks
        :param category: the category of collections to list
        :param ws_seen: IDs of workspaces already visited during recursion, used
          to break loops
        :param coll_seen: IDs of collections already produced, to remove
          duplicates when workspaces appear multiple times in
          the inheritance tree
        """
        if self.id in ws_seen:
            return
        if not self.can_display(pc):
            return
        ws_seen = (*ws_seen, self.id)
        if coll_seen is None:
            coll_seen = set()
        for collection in (
            self.collections.filter(category=category)
            .order_by("name")
            .select_related("workspace")
        ):
            if collection.id not in coll_seen:
                yield collection
                coll_seen.add(collection.id)
        for chain in self.chain_parents.select_related("parent").order_by(
            "order"
        ):
            yield from chain.parent.list_accessible_collections(
                pc, category, ws_seen=ws_seen, coll_seen=coll_seen
            )


class WorkspaceChain(models.Model):
    """Workspace chaining model."""

    child = models.ForeignKey(
        Workspace,
        on_delete=models.CASCADE,
        related_name="chain_parents",
        help_text="Workspace that falls back on `parent` for lookups",
    )
    parent = models.ForeignKey(
        Workspace,
        on_delete=models.CASCADE,
        related_name="chain_children",
        help_text="Workspace to be looked up if lookup in `child` fails",
    )
    order = models.IntegerField(
        help_text="Lookup order of this element in the chain",
    )

    class Meta(TypedModelMeta):
        constraints = [
            UniqueConstraint(
                fields=["child", "parent"],
                name="%(app_label)s_%(class)s_unique_child_parent",
            ),
            UniqueConstraint(
                fields=["child", "order"],
                name="%(app_label)s_%(class)s_unique_child_order",
            ),
        ]

    @override
    def __str__(self) -> str:
        """Return basic information of Workspace."""
        return f"{self.order}:{self.child.name}→{self.parent.name}"


class WorkspaceRole(models.Model):
    """Role assignments for workspaces."""

    Roles: TypeAlias = WorkspaceRoles

    resource = models.ForeignKey(
        Workspace,
        on_delete=models.CASCADE,
        related_name="roles",
    )

    group = models.ForeignKey(
        "Group",
        on_delete=models.CASCADE,
        related_name="workspace_roles",
    )

    role = models.CharField(max_length=16, choices=Roles.choices)

    class Meta(TypedModelMeta):
        constraints = [
            UniqueConstraint(
                fields=["resource", "group", "role"],
                name="%(app_label)s_%(class)s_unique_resource_group_role",
            ),
        ]

    @override
    def __str__(self) -> str:
        """Return a description of the role assignment."""
        return f"{self.group}─{self.role}⟶{self.resource}"



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Debusine command line interface, asset management commands."""

import argparse
from typing import override

from debusine.assets import AssetCategory, asset_data_model
from debusine.client.commands.base import (
    DebusineCommand,
    RequiredInputDataCommand,
    WorkspaceCommand,
)


class Create(RequiredInputDataCommand, WorkspaceCommand, group="asset"):
    """Create an asset."""

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            "category",
            type=AssetCategory,
            help="Category of asset",
        )

    @override
    def run(self) -> None:
        category = self.args.category

        asset_data = asset_data_model(category, self.input_data)

        with self._api_call_or_fail():
            asset = self.debusine.asset_create(
                category=category,
                data=asset_data,
                workspace=self.workspace,
            )

        # TODO: Ideally this would show a URL instead like everything
        # else, but assets don't have a web view at the moment.
        self.print_yaml_output(
            {
                "result": "success",
                "message": (
                    f"New asset created in {self.debusine.base_api_url} "
                    f"in workspace {asset.workspace} with id {asset.id}."
                ),
            }
        )


class LegacyCreate(
    Create, name="create-asset", deprecated="see `asset create`"
):
    """Create an asset."""


class List(DebusineCommand, group="asset"):
    """List assets."""

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            "--id",
            metavar="ID",
            type=int,
            help="Show an individual asset by ID.",
        )
        parser.add_argument(
            "--workspace",
            "-w",
            metavar="NAME",
            type=str,
            help="List all assets in workspace.",
        )
        parser.add_argument(
            "--work-request",
            metavar="ID",
            type=int,
            help="List assets created by work request.",
        )

    @override
    def run(self) -> None:
        asset = self.args.id
        workspace = self.args.workspace
        work_request = self.args.work_request
        if asset is None and workspace is None and work_request is None:
            self._fail(
                "At least one of --id, --workspace, and --work-request must "
                "be specified."
            )
        with self._api_call_or_fail():
            assets = self.debusine.asset_list(
                asset_id=asset,
                work_request=work_request,
                workspace=workspace,
            )

        self.print_yaml_output(assets.model_dump(mode="json"))


class LegacyList(List, name="list-assets", deprecated="see `asset list`"):
    """List assets."""



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command scope."""

from typing import ClassVar, override
from unittest import mock

from django.core.management import CommandError

from debusine.db.models import (
    DEFAULT_FILE_STORE_NAME,
    FileStore,
    Group,
    Scope,
    User,
)
from debusine.db.models.scopes import ScopeRole
from debusine.django.management.tests import call_command
from debusine.server.management.commands.scope import Command
from debusine.server.management.commands.tests.utils import TabularOutputTests
from debusine.test.django import TestCase


class CreateScopeCommandTests(TestCase):
    """Tests for `scope create` management command."""

    def assertHasScope(
        self,
        name: str = "scope",
        group_name: str | None = None,
        users: list[User] | None = None,
        label: str | None = None,
        icon: str | None = None,
        file_store_name: str = DEFAULT_FILE_STORE_NAME,
    ) -> None:
        """Check that the given scope exists, with its Owners group."""
        if users is None:
            users = []
        scope = Scope.objects.get(name=name)
        if group_name is not None:
            admin_group = Group.objects.get(scope=scope, name=group_name)
            self.assertQuerySetEqual(admin_group.users.all(), users)
            role = ScopeRole.objects.get(resource=scope, group=admin_group)
            self.assertEqual(role.role, ScopeRole.Roles.OWNER)
        if label is None:
            self.assertEqual(scope.label, scope.name.capitalize())
        else:
            self.assertEqual(scope.label, label)
        if icon is None:
            self.assertEqual(scope.icon, "")
        else:
            self.assertEqual(scope.icon, icon)
        self.assertQuerySetEqual(
            scope.file_stores.values_list(
                "name",
                "filestoreinscope__upload_priority",
                "filestoreinscope__download_priority",
            ),
            [(file_store_name, 100, 100)],
        )

    def test_create(self) -> None:
        """Test a successful create."""
        stdout, stderr, exit_code = call_command(
            "scope", "create", "scope", "--with-owners-group"
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        self.assertHasScope(group_name="Owners")

    def test_create_with_label(self) -> None:
        """Test setting label on create."""
        stdout, stderr, exit_code = call_command(
            "scope", "create", "scope", "--label=Scope Name"
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        self.assertHasScope(label="Scope Name")

    def test_update_label(self) -> None:
        """Test updating a label with create."""
        stdout, stderr, exit_code = call_command("scope", "create", "scope")
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        self.assertHasScope(label="Scope")

        stdout, stderr, exit_code = call_command(
            "scope", "create", "scope", "--label=Changed"
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        self.assertHasScope(label="Changed")

    def test_create_with_icon(self) -> None:
        """Test setting icon on create."""
        stdout, stderr, exit_code = call_command(
            "scope", "create", "scope", "--icon=web/icons/scope.svg"
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        self.assertHasScope(icon="web/icons/scope.svg")

    def test_update_icon(self) -> None:
        """Test updating a label with create."""
        stdout, stderr, exit_code = call_command("scope", "create", "scope")
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        self.assertHasScope()

        stdout, stderr, exit_code = call_command(
            "scope", "create", "scope", "--icon=web/icons/scope.svg"
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        self.assertHasScope(icon="web/icons/scope.svg")

    def test_create_custom_group_name(self) -> None:
        """Test a successful create with specified group name."""
        stdout, stderr, exit_code = call_command(
            "scope", "create", "scope", "--with-owners-group=Admin"
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        self.assertHasScope(group_name="Admin")

    def test_create_custom_file_store_name(self) -> None:
        """Test a successful create with specified file store name."""
        FileStore.objects.create(
            name="custom",
            backend=FileStore.BackendChoices.MEMORY,
            configuration={"name": "custom"},
        )

        stdout, stderr, exit_code = call_command(
            "scope", "create", "scope", "--file-store=custom"
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        self.assertHasScope(file_store_name="custom")

    def test_idempotent(self) -> None:
        """Test idempotence in creating twice."""
        call_command("scope", "create", "scope", "--with-owners-group")
        scope = Scope.objects.get(name="scope")
        admin_group = Group.objects.get(scope=scope, name="Owners")
        admin_group.users.set([self.playground.get_default_user()])

        # Recreate: the existing database objects are maintained
        call_command("scope", "create", "scope", "--with-owners-group")
        new_scope = Scope.objects.get(name="scope")
        self.assertEqual(new_scope.pk, scope.pk)
        new_group = Group.objects.get(scope=new_scope, name="Owners")
        self.assertEqual(new_group.pk, admin_group.pk)

        # The previous group membership is also maintained
        self.assertHasScope(users=[self.playground.get_default_user()])

    def test_no_admin_group(self) -> None:
        """Test a successful create with no admin group."""
        stdout, stderr, exit_code = call_command("scope", "create", "scope")
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        self.assertHasScope()
        self.assertFalse(Group.objects.filter(scope__name="scope").exists())

    def test_new_name_invalid(self) -> None:
        """Test creating with an invalid name."""
        stdout, stderr, exit_code = call_command("scope", "create", "a/b")
        self.assertEqual(stdout, "")
        self.assertEqual(
            stderr.splitlines(),
            [
                "New scope would be invalid:",
                "* name: 'a/b' is not a valid scope name",
            ],
        )
        self.assertEqual(exit_code, 3)
        self.assertFalse(Scope.objects.filter(name="a/b").exists())

    def test_file_store_does_not_exist(self) -> None:
        """Passing a nonexistent file store name results in an error."""
        with self.assertRaisesRegex(
            CommandError, r"File store 'nonexistent' not found"
        ) as exc:
            call_command("scope", "create", "scope", "--file-store=nonexistent")

        self.assertEqual(getattr(exc.exception, "returncode"), 3)


class RenameScopeCommandTests(TestCase):
    """Tests for `scope rename` management command."""

    scope1: ClassVar[Scope]
    scope2: ClassVar[Scope]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.scope1 = cls.playground.get_or_create_scope(name="scope1")
        cls.scope2 = cls.playground.get_or_create_scope(name="scope2")

    def test_rename(self) -> None:
        """Test a successful rename."""
        stdout, stderr, exit_code = call_command(
            "scope", "rename", "scope1", "newname"
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        self.scope1.refresh_from_db()
        self.assertEqual(self.scope1.name, "newname")

    def test_noop(self) -> None:
        """Test renaming to current name."""
        with mock.patch("debusine.db.models.Scope.save") as save:
            stdout, stderr, exit_code = call_command(
                "scope", "rename", "scope1", "scope1"
            )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        save.assert_not_called()

    def test_source_does_not_exist(self) -> None:
        """Test renaming a nonexisting scope."""
        with self.assertRaisesRegex(
            CommandError, r"Scope 'does-not-exist' not found"
        ) as exc:
            call_command("scope", "rename", "does-not-exist", "newname")

        self.assertEqual(getattr(exc.exception, "returncode"), 3)

    def test_target_exists(self) -> None:
        """Test renaming with a name already in use."""
        stdout, stderr, exit_code = call_command(
            "scope", "rename", "scope1", "scope2"
        )
        self.assertEqual(stdout, "")
        self.assertEqual(
            stderr.splitlines(),
            [
                "Renamed scope would be invalid:",
                "* name: Scope with this Name already exists.",
            ],
        )

        self.assertEqual(exit_code, 3)
        self.scope1.refresh_from_db()
        self.assertEqual(self.scope1.name, "scope1")

    def test_new_name_invalid(self) -> None:
        """Test renaming to an invalid name."""
        stdout, stderr, exit_code = call_command(
            "scope", "rename", "scope1", "api"
        )
        self.assertEqual(stdout, "")
        self.assertEqual(
            stderr.splitlines(),
            [
                "Renamed scope would be invalid:",
                "* name: 'api' is not a valid scope name",
            ],
        )
        self.assertEqual(exit_code, 3)
        self.scope1.refresh_from_db()
        self.assertEqual(self.scope1.name, "scope1")

    def test_invalid_action(self) -> None:
        """Test invoking an invalid subcommand."""
        with self.assertRaisesRegex(
            CommandError, r"invalid choice: 'does-not-exist'"
        ) as exc:
            call_command("scope", "does-not-exist")

        self.assertEqual(getattr(exc.exception, "returncode"), 1)

    def test_unexpected_action(self) -> None:
        """Test a subcommand with no implementation."""
        command = Command()

        with self.assertRaisesRegex(
            CommandError, r"Action 'does_not_exist' not found"
        ) as exc:
            command.handle(action="does_not_exist")

        self.assertEqual(getattr(exc.exception, "returncode"), 3)


class AddFileStoreScopeCommandTests(TestCase):
    """Tests for `scope add_file_store` management command."""

    scope: ClassVar[Scope]
    file_store: ClassVar[FileStore]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.scope = cls.playground.get_or_create_scope(name="scope")
        cls.file_store = FileStore.objects.create(
            name="store",
            backend=FileStore.BackendChoices.MEMORY,
            configuration={"name": "store"},
        )

    def test_defaults(self) -> None:
        """Add a file store to a scope, with default properties."""
        stdout, stderr, exit_code = call_command(
            "scope",
            "add_file_store",
            self.scope.name,
            self.file_store.name,
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        file_store_extra = self.scope.filestoreinscope_set.get(
            file_store=self.file_store
        )
        self.assertIsNone(file_store_extra.upload_priority)
        self.assertIsNone(file_store_extra.download_priority)
        self.assertFalse(file_store_extra.populate)
        self.assertFalse(file_store_extra.drain)
        self.assertIsNone(file_store_extra.drain_to)
        self.assertFalse(file_store_extra.read_only)
        self.assertFalse(file_store_extra.write_only)
        self.assertIsNone(file_store_extra.soft_max_size)

    def test_properties(self) -> None:
        """Add a file store to a scope, with non-default properties."""
        stdout, stderr, exit_code = call_command(
            "scope",
            "add_file_store",
            self.scope.name,
            self.file_store.name,
            "--upload-priority",
            "50",
            "--download-priority",
            "200",
            "--populate",
            "--no-drain",
            "--drain-to",
            "other_store",
            "--soft-max-size",
            "1048576",
            "--no-read-only",
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        file_store_extra = self.scope.filestoreinscope_set.get(
            file_store=self.file_store
        )
        self.assertEqual(file_store_extra.upload_priority, 50)
        self.assertEqual(file_store_extra.download_priority, 200)
        self.assertTrue(file_store_extra.populate)
        self.assertFalse(file_store_extra.drain)
        self.assertEqual(file_store_extra.drain_to, "other_store")
        self.assertFalse(file_store_extra.read_only)
        self.assertFalse(file_store_extra.write_only)
        self.assertEqual(file_store_extra.soft_max_size, 1048576)

    def test_nonexistent_scope(self) -> None:
        """Adding a file store to a nonexistent scope fails."""
        with self.assertRaisesRegex(
            CommandError, r"Scope 'does-not-exist' not found"
        ) as exc:
            call_command(
                "scope",
                "add_file_store",
                "does-not-exist",
                self.file_store.name,
            )

        self.assertEqual(getattr(exc.exception, "returncode"), 3)

    def test_nonexistent_file_store(self) -> None:
        """Adding a nonexistent file store to a scope fails."""
        with self.assertRaisesRegex(
            CommandError, r"File store 'does-not-exist' not found"
        ) as exc:
            call_command(
                "scope", "add_file_store", self.scope.name, "does-not-exist"
            )

        self.assertEqual(getattr(exc.exception, "returncode"), 3)


class EditFileStoreScopeCommandTests(TestCase):
    """Tests for `scope edit_file_store` management command."""

    scope: ClassVar[Scope]
    file_store: ClassVar[FileStore]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.scope = cls.playground.get_or_create_scope(name="scope")
        cls.file_store = FileStore.objects.create(
            name="store",
            backend=FileStore.BackendChoices.MEMORY,
            configuration={"name": "store"},
        )

    def test_set_all_properties(self) -> None:
        """Set all properties of a file store in a scope."""
        self.scope.file_stores.add(
            self.file_store,
            through_defaults={"populate": True, "write_only": True},
        )

        stdout, stderr, exit_code = call_command(
            "scope",
            "edit_file_store",
            self.scope.name,
            self.file_store.name,
            "--upload-priority",
            "50",
            "--download-priority",
            "200",
            "--no-populate",
            "--drain",
            "--drain-to",
            "other_store",
            "--read-only",
            "--no-write-only",
            "--soft-max-size",
            "1048576",
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        file_store_extra = self.scope.filestoreinscope_set.get(
            file_store=self.file_store
        )
        self.assertEqual(file_store_extra.upload_priority, 50)
        self.assertEqual(file_store_extra.download_priority, 200)
        self.assertFalse(file_store_extra.populate)
        self.assertTrue(file_store_extra.drain)
        self.assertEqual(file_store_extra.drain_to, "other_store")
        self.assertTrue(file_store_extra.read_only)
        self.assertFalse(file_store_extra.write_only)
        self.assertEqual(file_store_extra.soft_max_size, 1048576)

    def test_unchanged(self) -> None:
        """
        Leave all properties of a file store in a scope unchanged.

        This is mainly here for ease of test coverage.
        """
        self.scope.file_stores.add(
            self.file_store,
            through_defaults={
                "upload_priority": 100,
                "download_priority": 100,
                "populate": True,
                "drain": False,
                "drain_to": "other_store",
                "read_only": False,
                "write_only": True,
                "soft_max_size": 1048576,
            },
        )
        stdout, stderr, exit_code = call_command(
            "scope", "edit_file_store", self.scope.name, self.file_store.name
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        file_store_extra = self.scope.filestoreinscope_set.get(
            file_store=self.file_store
        )
        self.assertEqual(file_store_extra.upload_priority, 100)
        self.assertEqual(file_store_extra.download_priority, 100)
        self.assertTrue(file_store_extra.populate)
        self.assertFalse(file_store_extra.drain)
        self.assertEqual(file_store_extra.drain_to, "other_store")
        self.assertFalse(file_store_extra.read_only)
        self.assertTrue(file_store_extra.write_only)
        self.assertEqual(file_store_extra.soft_max_size, 1048576)

    def test_unset_upload_priority(self) -> None:
        """Unset upload priority for a file store in a scope."""
        self.scope.file_stores.add(
            self.file_store,
            through_defaults={"upload_priority": 50},
        )

        stdout, stderr, exit_code = call_command(
            "scope",
            "edit_file_store",
            self.scope.name,
            self.file_store.name,
            "--upload-priority",
            "",
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        file_store_extra = self.scope.filestoreinscope_set.get(
            file_store=self.file_store
        )
        self.assertIsNone(file_store_extra.upload_priority)

    def test_unset_download_priority(self) -> None:
        """Unset download priority for a file store in a scope."""
        self.scope.file_stores.add(
            self.file_store, through_defaults={"download_priority": 50}
        )

        stdout, stderr, exit_code = call_command(
            "scope",
            "edit_file_store",
            self.scope.name,
            self.file_store.name,
            "--download-priority",
            "",
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        file_store_extra = self.scope.filestoreinscope_set.get(
            file_store=self.file_store
        )
        self.assertIsNone(file_store_extra.download_priority)

    def test_unset_soft_max_size(self) -> None:
        """Unset soft maximum size for a file store in a scope."""
        self.scope.file_stores.add(
            self.file_store, through_defaults={"soft_max_size": 1048576}
        )

        stdout, stderr, exit_code = call_command(
            "scope",
            "edit_file_store",
            self.scope.name,
            self.file_store.name,
            "--soft-max-size",
            "",
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        file_store_extra = self.scope.filestoreinscope_set.get(
            file_store=self.file_store
        )
        self.assertIsNone(file_store_extra.soft_max_size)

    def test_nonexistent_scope(self) -> None:
        """Editing a file store in a nonexistent scope fails."""
        with self.assertRaisesRegex(
            CommandError, r"Scope 'does-not-exist' not found"
        ) as exc:
            call_command(
                "scope",
                "edit_file_store",
                "does-not-exist",
                self.file_store.name,
            )

        self.assertEqual(getattr(exc.exception, "returncode"), 3)

    def test_nonexistent_file_store(self) -> None:
        """Editing a nonexistent file store in a scope fails."""
        with self.assertRaisesRegex(
            CommandError,
            rf"Scope '{self.scope.name}' has no file store named "
            rf"'does-not-exist'",
        ) as exc:
            call_command(
                "scope", "edit_file_store", self.scope.name, "does-not-exist"
            )

        self.assertEqual(getattr(exc.exception, "returncode"), 3)

    def test_not_in_this_scope(self) -> None:
        """Editing a file store that is not in the given scope fails."""
        with self.assertRaisesRegex(
            CommandError,
            rf"Scope '{self.scope.name}' has no file store named "
            rf"'{self.file_store.name}'",
        ) as exc:
            call_command(
                "scope",
                "edit_file_store",
                self.scope.name,
                self.file_store.name,
            )

        self.assertEqual(getattr(exc.exception, "returncode"), 3)


class RemoveFileStoreScopeCommandTests(TestCase):
    """Tests for `scope remove_file_store` management command."""

    scope: ClassVar[Scope]
    file_store: ClassVar[FileStore]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.scope = cls.playground.get_or_create_scope(name="scope")
        cls.file_store = FileStore.objects.create(
            name="store",
            backend=FileStore.BackendChoices.MEMORY,
            configuration={"name": "store"},
        )

    def test_success(self) -> None:
        """Remove a file store from a scope."""
        self.scope.file_stores.set([self.file_store])
        workspace = self.playground.create_workspace(scope=self.scope)
        artifact, _ = self.playground.create_artifact(
            paths=["test"], workspace=workspace, create_files=True
        )
        other_store = FileStore.objects.create(
            name="other",
            backend=FileStore.BackendChoices.MEMORY,
            configuration={"name": "other"},
        )
        self.scope.file_stores.add(other_store)
        other_store.files.add(artifact.files.get())

        stdout, stderr, exit_code = call_command(
            "scope", "remove_file_store", self.scope.name, self.file_store.name
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        self.assertNotIn(self.file_store, self.scope.file_stores.all())

    def test_remove_file_store_with_unique_files(self) -> None:
        """Cannot remove a store if some files in this scope are only there."""
        self.scope.file_stores.set([self.file_store])
        workspace = self.playground.create_workspace(scope=self.scope)
        self.playground.create_artifact(
            paths=["test"], workspace=workspace, create_files=True
        )

        with self.assertRaisesRegex(
            CommandError,
            rf"Scope '{self.scope.name}' contains some files that are only in "
            rf"file store '{self.file_store.name}'; drain it first",
        ) as exc:
            call_command(
                "scope",
                "remove_file_store",
                self.scope.name,
                self.file_store.name,
            )

        self.assertEqual(getattr(exc.exception, "returncode"), 3)

    def test_remove_file_store_nonexistent_scope(self) -> None:
        """Removing a file store from a nonexistent scope fails."""
        with self.assertRaisesRegex(
            CommandError, r"Scope 'does-not-exist' not found"
        ) as exc:
            call_command(
                "scope",
                "remove_file_store",
                "does-not-exist",
                self.file_store.name,
            )

        self.assertEqual(getattr(exc.exception, "returncode"), 3)

    def test_remove_file_store_nonexistent_file_store(self) -> None:
        """Removing a nonexistent file store from a scope fails."""
        with self.assertRaisesRegex(
            CommandError,
            rf"Scope '{self.scope.name}' has no file store named "
            rf"'does-not-exist'",
        ) as exc:
            call_command(
                "scope", "remove_file_store", self.scope.name, "does-not-exist"
            )

        self.assertEqual(getattr(exc.exception, "returncode"), 3)

    def test_remove_file_store_not_in_this_scope(self) -> None:
        """Removing a file store that is not in the given scope fails."""
        with self.assertRaisesRegex(
            CommandError,
            rf"Scope '{self.scope.name}' has no file store named "
            rf"'{self.file_store.name}'",
        ) as exc:
            call_command(
                "scope",
                "remove_file_store",
                self.scope.name,
                self.file_store.name,
            )

        self.assertEqual(getattr(exc.exception, "returncode"), 3)


class ShowScopeCommandTests(TabularOutputTests, TestCase):
    """Tests for `scope show` management command."""

    file_stores: ClassVar[list[FileStore]]
    scope: ClassVar[Scope]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.file_stores = [
            FileStore.objects.create(
                name=name, backend=backend, configuration=configuration
            )
            for name, backend, configuration in (
                (
                    "default",
                    FileStore.BackendChoices.MEMORY,
                    {"name": "default"},
                ),
                ("backup", FileStore.BackendChoices.MEMORY, {"name": "backup"}),
                ("fast_download", FileStore.BackendChoices.LOCAL, {}),
            )
        ]
        cls.scope = cls.playground.get_or_create_scope(
            name="scope", label="Label"
        )
        cls.scope.file_stores.clear()
        cls.scope.file_stores.add(
            cls.file_stores[0], through_defaults={"upload_priority": 100}
        )
        cls.scope.file_stores.add(cls.file_stores[1])
        cls.scope.file_stores.add(
            cls.file_stores[2], through_defaults={"download_priority": 200}
        )

    def test_show(self) -> None:
        """Show a scope."""
        workspace = self.playground.create_workspace(scope=self.scope)
        artifacts = [
            self.playground.create_artifact(
                paths=["test"], workspace=workspace, create_files=True
            )[0]
            for _ in range(5)
        ]
        for artifact in artifacts[:2]:
            self.file_stores[1].files.add(artifact.files.get())

        with self.assertPrintsTable() as output:
            call_command("scope", "show", self.scope.name)

        self.assertEqual(output.col(0), ["scope", "", ""])
        self.assertEqual(output.col(1), ["Label", "", ""])
        self.assertEqual(output.col(2), ["default", "fast_download", "backup"])
        self.assertEqual(output.col(3), ["Memory", "Local", "Memory"])
        self.assertEqual(output.col(4), ["100", "", ""])
        self.assertEqual(output.col(5), ["", "200", ""])
        self.assertEqual(output.col(6), ["5", "0", "2"])

    def test_show_duplicate_files(self) -> None:
        """Count the number of individual (distinct) files."""
        workspace = self.playground.create_workspace(scope=self.scope)
        # 3 files with the same contents are stored as a single File
        for _ in range(3):
            self.playground.create_artifact(
                paths={"test": b"static data"},
                workspace=workspace,
                create_files=True,
            )
        with self.assertPrintsTable() as output:
            call_command("scope", "show", self.scope.name)

        self.assertEqual(output.col(0), ["scope", "", ""])
        self.assertEqual(output.col(1), ["Label", "", ""])
        self.assertEqual(output.col(2), ["default", "fast_download", "backup"])
        self.assertEqual(output.col(3), ["Memory", "Local", "Memory"])
        self.assertEqual(output.col(4), ["100", "", ""])
        self.assertEqual(output.col(5), ["", "200", ""])
        self.assertEqual(output.col(6), ["1", "0", "0"])



#!/usr/bin/env python3

# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Debusine: Interacts with debusine server.

Debusine fetches information, submit work requests and other operations.
"""

import asyncio
import datetime as dt
import io
import json
import logging
import os
import os.path
import shutil
import subprocess
import tempfile
from collections.abc import Coroutine, Iterable, Sequence
from pathlib import Path, PurePath
from typing import Any, Literal, Protocol, TYPE_CHECKING, overload
from urllib.parse import quote, urlencode, urlparse, urlunparse

import aiohttp
import magic
import pydantic
import requests
from requests_toolbelt.downloadutils.stream import stream_response_to_file

from debusine.artifacts import LocalArtifact
from debusine.artifacts.models import (
    ArtifactData,
    CollectionCategory,
    WorkRequestResults,
)
from debusine.assets import AssetCategory, BaseAssetDataModel
from debusine.client.debusine_http_client import DebusineHttpClient
from debusine.client.exceptions import TokenDisabledError
from debusine.client.file_uploader import FileUploader
from debusine.client.models import (
    ArtifactCreateRequest,
    ArtifactResponse,
    AssetCreateRequest,
    AssetPermissionCheckResponse,
    AssetResponse,
    AssetsResponse,
    CollectionData,
    CollectionDataNew,
    CollectionIdentifierRequest,
    CollectionRelationResponse,
    CollectionRelationSetRequest,
    CollectionRelationSetResponse,
    CollectionRelationType,
    CollectionRoles,
    CollectionTokenGrantRequest,
    CollectionTokenGrantResponse,
    CreateWorkflowRequest,
    EmitMetricRequest,
    FileRequest,
    FilesRequestType,
    LookupChildType,
    LookupMultiple,
    LookupMultipleRequest,
    LookupMultipleResponse,
    LookupSingle,
    LookupSingleRequest,
    LookupSingleResponse,
    LookupSingleResponseArtifact,
    LookupSingleResponseCollection,
    OnWorkRequestCompleted,
    RelationCreateRequest,
    RelationResponse,
    RelationType,
    RelationsResponse,
    RemoteArtifact,
    TaskConfigurationCollectionContents,
    TaskConfigurationCollectionUpdateResults,
    WorkRequestExternalDebsignRequest,
    WorkRequestRequest,
    WorkRequestResponse,
    WorkflowTemplateData,
    WorkflowTemplateDataNew,
    WorkspaceInheritanceChain,
)
from debusine.utils import get_user_agent
from debusine.utils.urls import DEFAULT_REQUEST_TIMEOUT, URLPath

if TYPE_CHECKING:
    from debusine.client.server_notifications import ServerNotifications
    from debusine.client.task_configuration import (
        LocalTaskConfigurationRepository,
        RemoteTaskConfigurationRepository,
    )
    from debusine.tasks.models import OutputData


class ServerNotificationDecodeError(Exception):
    """Error in decoding server notifications."""


class _MessageProcessor(Protocol):
    """A callable that processes a message from the server."""

    def __call__(
        self, *, msg_content: dict[str, Any]
    ) -> Coroutine[Any, Any, None] | None: ...


def serialize_local_artifact(
    artifact: LocalArtifact[Any],
    *,
    workspace: str | None,
    work_request: int | None = None,
    expire_at: dt.datetime | None = None,
) -> dict[str, Any]:
    """Return dictionary to be used by the API to create an artifact."""
    files: FilesRequestType = FilesRequestType({})
    for artifact_path, local_path in artifact.files.items():
        content_type = artifact.content_types.get(artifact_path)
        if content_type is None:
            match PurePath(artifact_path).suffix:
                case ".buildinfo" | ".dsc" | ".changes":
                    # Policy requires these to be UTF-8.
                    content_type = "text/plain; charset=utf-8"
                case ".md":
                    # libmagic guesses this as text/plain.
                    content_type = "text/markdown; charset=utf-8"
                case _:
                    # Try to automatically detect a suitable Content-Type.
                    # Note that for the moment we deliberately don't look at
                    # the content of compressed files, since as a general
                    # rule we don't want browsers to automatically
                    # uncompress them.
                    #
                    # We run Content-Type detection on the client so that
                    # the server doesn't need to run possibly
                    # security-sensitive parsing/guessing code, but we also
                    # don't want to give the client entirely free rein since
                    # some Content-Types could be used to attack browsers.
                    # The server therefore squashes Content-Types that it
                    # doesn't know to be safe; see debusine.web.views.files.
                    content_type = magic.Magic(
                        mime=True, mime_encoding=True
                    ).from_file(local_path)
        files[artifact_path] = FileRequest.create_from(
            local_path, content_type=content_type
        )
    for artifact_path, file_request in artifact.remote_files.items():
        files[artifact_path] = file_request

    artifact.model_validate(artifact)

    serialized = ArtifactCreateRequest(
        workspace=workspace,
        category=artifact.category,
        files=files,
        data=(
            artifact.data.model_dump()
            if isinstance(artifact.data, ArtifactData)
            else artifact.data
        ),
        work_request=work_request,
        expire_at=expire_at,
    ).model_dump(mode="json")

    # If the workspace was not specified: do not send it to the API.
    # The server will assign it.
    if serialized["workspace"] is None:
        del serialized["workspace"]

    return serialized


class Debusine:
    """Class to interact with debusine server."""

    API_VERSION = '1.0'

    def __init__(
        self,
        base_api_url: str,
        api_token: str | None = None,
        scope: str | None = None,
        *,
        logger: logging.Logger,
        timeout: int | tuple[int, int] = DEFAULT_REQUEST_TIMEOUT,
    ) -> None:
        """
        Initialize client.

        :param base_api_url: URL for API endpoint (e.g. http://localhost/api)
        :param api_token: optional token to be used for the calls.
        :param scope: scope to be used for the calls.
        :param logger: logger to be used for the calls.
        :param timeout: timeout for HTTP requests (connect, read).
        """
        self.base_api_url: str = base_api_url
        self.token: str | None = api_token
        self.api_url = base_api_url.rstrip("/") + "/" + self.API_VERSION

        # TODO: Currently Workers require the use of an unscoped client for
        # their websocket connection.
        # Ideally 'scope' will become a constructor parameter, where
        # self.scope can't be None, simplifying/hardening the code.
        self.scope: str | None = scope

        self._logger = logger

        self.timeout = timeout

        self._debusine_http_client: DebusineHttpClient = DebusineHttpClient(
            self.api_url, api_token, scope, timeout=self.timeout
        )

    def webui_url(self, path: str) -> str:
        """Generate a URL to the web UI given a full path."""
        parsed = urlparse(self.base_api_url)
        parsed_path = URLPath(parsed.path)
        if parsed_path and parsed_path[-1] == "api":
            parsed_path.pop()
        parsed_path += path
        parsed = parsed._replace(path=str(parsed_path))
        return urlunparse(parsed)

    def work_request_iter(self) -> Iterable[WorkRequestResponse]:
        """
        List all WorkRequests.

        :raises many: see _api_request method documentation.
        """
        return self._debusine_http_client.iter_paginated_get(
            "/work-request/", WorkRequestResponse
        )

    def work_request_get(self, work_request_id: int) -> WorkRequestResponse:
        """
        Get WorkRequest for work_request_id.

        :param work_request_id: id to fetch the status of.
        :raises many: see _api_request method documentation.
        """
        return self._debusine_http_client.get(
            f"/work-request/{work_request_id}/", WorkRequestResponse
        )

    def work_request_update(
        self, work_request_id: int, *, priority_adjustment: int
    ) -> WorkRequestResponse:
        """Update properties of work_request_id."""
        return self._debusine_http_client.patch(
            path=f"/work-request/{work_request_id}/",
            data={"priority_adjustment": priority_adjustment},
            expected_class=WorkRequestResponse,
        )

    def work_request_retry(
        self,
        work_request_id: int,
    ) -> WorkRequestResponse:
        """Retry a work request."""
        return self._debusine_http_client.post(
            path=f"/work-request/{work_request_id}/retry/",
            data={},
            expected_class=WorkRequestResponse,
        )

    def work_request_abort(
        self,
        work_request_id: int,
    ) -> WorkRequestResponse:
        """Abort a work request."""
        return self._debusine_http_client.post(
            path=f"/work-request/{work_request_id}/abort/",
            data={},
            expected_class=WorkRequestResponse,
        )

    @staticmethod
    def work_request_completed_path(work_request_id: int) -> str:
        """Return path to update the completed result for work_request_id."""
        return f"/work-request/{work_request_id}/completed/"

    def work_request_completed_update(
        self,
        work_request_id: int,
        result: WorkRequestResults,
        output_data: "OutputData",
    ) -> None:
        """Update work_request_id as completed with result."""
        return self._debusine_http_client.put(
            path=self.work_request_completed_path(work_request_id),
            data={
                "result": str(result),
                "output_data": output_data.model_dump(
                    mode="json", exclude_unset=True
                ),
            },
            expected_class=None,
        )

    def work_request_create(
        self, work_request: WorkRequestRequest
    ) -> WorkRequestResponse:
        """
        Create a work request (via POST /work-request/).

        :return: WorkRequest returned by the server.
        :raises: see _api_request method documentation.
        """
        data = work_request.model_dump()

        # If workspace is None: the client does not want to specify
        # a workspace. Do not send it to the server: the server will
        # assign the default one
        if data["workspace"] is None:
            del data["workspace"]

        return self._debusine_http_client.post(
            "/work-request/",
            WorkRequestResponse,
            data=data,
        )

    def work_request_external_debsign_get(
        self, work_request_id: int
    ) -> WorkRequestResponse:
        """Get data about work request waiting for an external `debsign`."""
        return self._debusine_http_client.get(
            f"/work-request/{work_request_id}/external-debsign/",
            WorkRequestResponse,
        )

    def work_request_external_debsign_complete(
        self,
        work_request_id: int,
        external_debsign_request: WorkRequestExternalDebsignRequest,
    ) -> WorkRequestResponse:
        """Provide external `debsign` data to a work request."""
        data = external_debsign_request.model_dump()

        return self._debusine_http_client.post(
            f"/work-request/{work_request_id}/external-debsign/",
            WorkRequestResponse,
            data,
        )

    def workflow_template_create(
        self, workspace: str, workflow_template: WorkflowTemplateDataNew
    ) -> WorkflowTemplateData:
        """
        Create a workflow template (via POST /workflow-template/{workspace}/).

        :return: WorkflowTemplate returned by the server.
        :raises: see _api_request method documentation.
        """
        return self._debusine_http_client.post(
            f"/workflow-template/{quote(workspace)}/",
            WorkflowTemplateData,
            data=workflow_template.model_dump(),
            expected_statuses=[requests.codes.created],
        )

    def workflow_template_iter(
        self, workspace: str
    ) -> Iterable[WorkflowTemplateData]:
        """List workflow templates on a workspace."""
        return self._debusine_http_client.iter_paginated_get(
            f"/workflow-template/{quote(workspace)}/", WorkflowTemplateData
        )

    def workflow_template_get(
        self, workspace: str, workflow_template: int | str
    ) -> WorkflowTemplateData:
        """Fetch information about a workflow template."""
        name = quote(str(workflow_template))
        return self._debusine_http_client.get(
            f"/workflow-template/{quote(workspace)}/{name}/",
            expected_class=WorkflowTemplateData,
        )

    def workflow_template_update(
        self, workspace: str, data: WorkflowTemplateData
    ) -> WorkflowTemplateData:
        """Set information for a workflow template."""
        return self._debusine_http_client.patch(
            f"/workflow-template/{quote(workspace)}/{data.id}/",
            data=data.model_dump(mode="json"),
            expected_class=WorkflowTemplateData,
        )

    def workflow_template_delete(
        self, workspace: str, workflow_template: int | str
    ) -> None:
        """Set information for a workflow template."""
        name = str(workflow_template)
        return self._debusine_http_client.delete(
            f"/workflow-template/{quote(workspace)}/{quote(name)}/",
            expected_statuses=[requests.codes.no_content],
        )

    def workflow_create(
        self, create_workflow: CreateWorkflowRequest
    ) -> WorkRequestResponse:
        """
        Create a workflow (via POST /workflow/).

        :return: Workflow WorkRequest returned by the server.
        :raises: see _api_request method documentation.
        """
        data = create_workflow.model_dump()

        # If workspace is None: the client does not want to specify a
        # workspace. Do not send it to the server: the server will assign
        # the default one.
        if data["workspace"] is None:
            del data["workspace"]

        return self._debusine_http_client.post(
            "/workflow/",
            WorkRequestResponse,
            data=data,
            expected_statuses=[requests.codes.created],
        )

    def artifact_get(
        self, artifact_id: int, workspace: str | None = None
    ) -> ArtifactResponse:
        """
        Get artifact information.

        Use download_artifact() to download the artifact.
        """
        if workspace:
            path = f"/artifact/{quote(workspace)}/{artifact_id}/"
        else:
            path = f"/artifact/{artifact_id}/"
        return self._debusine_http_client.get(path, ArtifactResponse)

    def artifact_create(
        self,
        artifact: LocalArtifact[Any],
        *,
        workspace: str | None,
        work_request: int | None = None,
        expire_at: dt.datetime | None = None,
    ) -> ArtifactResponse:
        """Create artifact in the debusine server."""
        return self._debusine_http_client.post(
            "/artifact/",
            ArtifactResponse,
            data=serialize_local_artifact(
                artifact=artifact,
                workspace=workspace,
                work_request=work_request,
                expire_at=expire_at,
            ),
            expected_statuses=[requests.codes.created],
        )

    def _url_for_file_in_artifact(
        self, artifact_id: int, path_in_artifact: str
    ) -> str:
        """Return URL to upload a file in the artifact."""
        quoted_path = quote(path_in_artifact)

        return f"{self.api_url}/artifact/{artifact_id}/files/{quoted_path}/"

    def upload_files(
        self,
        artifact_id: int,
        upload_files: dict[str, Path],
        base_directory: Path | None = None,
    ) -> None:
        """
        Upload into artifact the files.

        :param artifact_id: artifact_id to upload files to.
        :param upload_files: list of files to upload.
        :param base_directory: base directory for relative path's files
         to upload.
        """
        # The file uploader won't work as an anonymous user.  If this
        # assertion is optimized out then the file upload will fail
        # harmlessly; this just helps with type-checking.
        assert self.token is not None
        assert self.scope is not None

        file_uploader = FileUploader(
            self.token,
            scope=self.scope,
            logger=self._logger,
            timeout=self.timeout,
        )

        for artifact_path, local_path in upload_files.items():
            if not local_path.is_absolute():
                if base_directory is None:
                    raise ValueError(
                        f"{local_path} is relative: base_directory "
                        "parameter cannot be None"
                    )

                local_path = base_directory.joinpath(local_path)

            url = self._url_for_file_in_artifact(artifact_id, artifact_path)

            file_uploader.upload(local_path, url)

    def upload_artifact(
        self,
        local_artifact: LocalArtifact[Any],
        *,
        workspace: str | None,
        work_request: int | None = None,
        expire_at: dt.datetime | None = None,
    ) -> RemoteArtifact:
        """Upload (create and upload files) the local_artifact to the server."""
        artifact_response = self.artifact_create(
            local_artifact,
            workspace=workspace,
            work_request=work_request,
            expire_at=expire_at,
        )

        files_to_upload: dict[str, Path] = {}
        for file_path_to_upload in artifact_response.files_to_upload:
            files_to_upload[file_path_to_upload] = local_artifact.files[
                file_path_to_upload
            ]

        self.upload_files(
            artifact_response.id,
            files_to_upload,
        )

        return RemoteArtifact(
            id=artifact_response.id,
            url=artifact_response.url,
            workspace=artifact_response.workspace,
        )

    def relation_create(
        self,
        artifact_id: int,
        target_id: int,
        relation_type: RelationType,
    ) -> RelationResponse:
        """
        Create a new relation between artifacts.

        :param artifact_id: relation from
        :param target_id: relation to
        :param relation_type: type of relation such as extends, relates-to,
          built-using
        :return: True if the relation already existed/has been created,
          False if it could not be created
        """
        relation_request = RelationCreateRequest(
            artifact=artifact_id, target=target_id, type=relation_type
        )
        return self._debusine_http_client.post(
            "/artifact-relation/",
            RelationResponse,
            data=relation_request.model_dump(
                include={"artifact", "target", "type"}
            ),
            expected_statuses=[requests.codes.ok, requests.codes.created],
        )

    def relation_list(
        self, *, artifact_id: int | None = None, target_id: int | None = None
    ) -> RelationsResponse:
        """
        List relations associated with an artifact.

        Exactly one of artifact_id or target_id must be set.

        :param artifact_id: search for relations from this artifact
        :param target_id: search for relations to this artifact
        :return: a list of relations
        :raises ValueError: if exactly one of artifact_id or target_id is not
          set
        """
        if (artifact_id is None) == (target_id is None):
            raise ValueError(
                "Exactly one of artifact_id or target_id must be set"
            )
        params = {}
        if artifact_id is not None:
            params["artifact"] = str(artifact_id)
        else:
            assert target_id is not None
            params["target_artifact"] = str(target_id)
        return self._debusine_http_client.get(
            "/artifact-relation/?" + urlencode(params), RelationsResponse
        )

    def collection_relation_list(
        self,
        workspace: str,
        *,
        source: str | None = None,
        target: str | None = None,
        relation_type: CollectionRelationType | None = None,
    ) -> Iterable[CollectionRelationResponse]:
        """
        List relations between collections in workspace.

        :param source: search for relations from this collection
        :param target: search for relations to this collection
        :param relation_type: search for this type of relation
        :return: a list of relations
        """
        params: dict[str, str] = {}
        if source is not None:
            params["source"] = source
        if target is not None:
            params["target"] = target
        if relation_type is not None:
            params["type"] = str(relation_type)

        endpoint = f"/collection-relation/{quote(workspace)}/"
        if params:
            endpoint += "?" + urlencode(params)
        return self._debusine_http_client.iter_paginated_get(
            endpoint, CollectionRelationResponse
        )

    def collection_relation_list_for_edit(
        self,
        workspace: str,
        *,
        source: str,
        relation_type: CollectionRelationType,
    ) -> tuple[list[CollectionRelationResponse], str]:
        """
        List relation targets to edit and return the list ETag.

        :param source: relation source collection, in name@category format
        :param relation_type: relation type to edit
        :return: (list of relations, ETag)
        """
        endpoint = f"/collection-relation/{quote(workspace)}/?" + urlencode(
            {"source": source, "type": str(relation_type)}
        )

        relations, headers = (
            self._debusine_http_client.get_single_page_with_headers(
                endpoint, CollectionRelationResponse
            )
        )

        return relations, headers["ETag"]

    def collection_relation_set(
        self,
        workspace: str,
        *,
        source: str,
        relation_type: CollectionRelationType,
        targets: list[CollectionIdentifierRequest],
        if_match: str,
    ) -> CollectionRelationSetResponse:
        """
        Replace all relation targets for a source collection and relation type.

        :param source: relation source collection, in name@category format
        :param relation_type: relation type to edit
        :param targets: full ordered relation target list (name@category)
        :param if_match: set the "If-Match" header to it
        :return: resulting ordered target list
        """
        response = self._debusine_http_client.put(
            f"/collection-relation/{quote(workspace)}/"
            f"{quote(source)}/{quote(relation_type)}/",
            CollectionRelationSetResponse,
            data=CollectionRelationSetRequest(targets=targets).model_dump(
                mode="json"
            ),
            expected_statuses=[requests.codes.ok],
            if_match=if_match,
        )
        assert response is not None
        return response

    @staticmethod
    def _should_download_as_tarball(artifact: ArtifactResponse) -> bool:
        """
        Decide whether to download artifact as a tarball or individually.

        Apply some heuristics to decide whether it's worth having the server
        build a tarball or not.
        """
        if len(artifact.files) == 1:
            return False
        total_size = 0
        for file in artifact.files.values():
            total_size += file.size
            if file.size > 20 * 1024 * 1024:
                return False
        return total_size < 100 * 1024 * 1024

    def _streaming_download(self, url: pydantic.AnyUrl) -> requests.Response:
        """Streaming download of url."""
        headers = {"Token": self.token, "User-Agent": get_user_agent()}
        response = requests.get(str(url), headers=headers, stream=True)
        response.raise_for_status()
        return response

    def download_artifact(
        self,
        artifact_id: int,
        destination: Path,
        *,
        tarball: bool = False,
        workspace: str | None = None,
    ) -> ArtifactResponse:
        """
        Download artifact_id into destination directory.

        :param artifact_id: artifact id to download
        :param destination: destination directory to download/uncompress
        :param tarball: True to only download the tarball (artifact-id.tar.gz),
          False to uncompress it
        """
        artifact = self.artifact_get(artifact_id, workspace=workspace)

        if not tarball and not self._should_download_as_tarball(artifact):
            self._logger.info("Downloading artifact files into %s", destination)
            for name, file in artifact.files.items():
                self._logger.debug(
                    "Downloading artifact file %s from %s", name, file.url
                )
                path = destination / name
                if path.parent != destination:
                    if not path.is_relative_to(destination):
                        raise AssertionError(f"Invalid filename {name}")
                    path.parent.mkdir(parents=True, exist_ok=True)
                stream_response_to_file(
                    self._streaming_download(file.url), path
                )
            return artifact

        url = artifact.download_tar_gz_url
        if tarball:
            destination_file = destination / f"artifact-{artifact_id}.tar.gz"
            stream_response_to_file(
                self._streaming_download(url),
                destination_file,
            )
            self._logger.info("Artifact downloaded: %s", destination_file)
            return artifact

        self._logger.info(
            "Downloading artifact and uncompressing into %s", destination
        )
        raw_response = self._streaming_download(url).raw
        raw_response.decode_content = True
        self._uncompress(raw_response, destination, url)

        return artifact

    def _uncompress(
        self,
        raw_response: io.IOBase,
        destination: Path,
        url: pydantic.AnyUrl,
    ) -> None:
        with (
            tempfile.TemporaryFile() as stderr_file,
            tempfile.TemporaryFile() as list_file,
        ):
            proc = subprocess.Popen(
                ["tar", "-xzv"],
                stdin=subprocess.PIPE,
                stderr=stderr_file,
                stdout=list_file,
                cwd=destination,
            )

            assert proc.stdin is not None
            shutil.copyfileobj(raw_response, proc.stdin)

            proc.stdin.close()

            returncode = proc.wait()

            list_file.seek(0)
            self._logger.info(
                list_file.read().decode(errors="ignore").rstrip("\n")
            )

            if returncode != 0:
                stderr_file.seek(0)
                stderr_contents = stderr_file.read().decode(errors="ignore")
                raise RuntimeError(
                    f"Error untarring {url}. "
                    f"Returncode: {returncode} stderr:\n{stderr_contents}"
                )

    def download_artifact_file(
        self,
        artifact_id: int | ArtifactResponse,
        path_in_artifact: str,
        destination: Path,
        workspace: str | None = None,
    ) -> ArtifactResponse:
        """
        Download a single file from artifact into destination directory.

        :param artifact_id: artifact id to download, or an
          :py:class:`ArtifactResponse` representing the artifact
        :param path_in_artifact: download the file from the artifact with
          this name
        :param destination: destination file to download to
        """
        if isinstance(artifact_id, ArtifactResponse):
            artifact = artifact_id
        else:
            artifact = self.artifact_get(artifact_id, workspace=workspace)

        if path_in_artifact not in artifact.files:
            raise ValueError(
                f"No file '{path_in_artifact}' in artifact {artifact.id}"
            )
        url = artifact.files[path_in_artifact].url
        response = self._streaming_download(url)
        stream_response_to_file(response, destination)
        self._logger.info("Artifact file downloaded: %s", destination)

        return artifact

    def asset_create(
        self,
        category: AssetCategory,
        data: BaseAssetDataModel,
        *,
        workspace: str,
        work_request: int | None = None,
    ) -> AssetResponse:
        """Create asset in the debusine server."""
        return self._debusine_http_client.post(
            "/asset/",
            AssetResponse,
            data=AssetCreateRequest(
                category=category,
                workspace=workspace,
                work_request=work_request,
                data=data.model_dump(),
            ).model_dump(),
            expected_statuses=[requests.codes.created],
        )

    def asset_list(
        self,
        *,
        asset_id: int | None = None,
        work_request: int | None = None,
        workspace: str | None = None,
    ) -> AssetsResponse:
        """Create asset in the debusine server."""
        params: dict[str, str | int] = {}
        if asset_id:
            params["asset"] = asset_id
        if work_request:
            params["work_request"] = work_request
        if workspace:
            params["workspace"] = workspace
        if not params:
            raise ValueError(
                "At least one of asset, work_request, and workspace must be "
                "specified"
            )
        query = urlencode(params)
        return self._debusine_http_client.get(
            f"/asset/?{query}",
            AssetsResponse,
        )

    def asset_permission_check(
        self,
        *,
        asset_category: str,
        asset_slug: str,
        permission_name: str,
        artifact_id: int,
        work_request_id: int,
        workspace: str,
    ) -> AssetPermissionCheckResponse:
        """Check work_request's permission to use an asset on an artifact."""
        return self._debusine_http_client.post(
            f"/asset/{quote(asset_category)}/{quote(asset_slug)}/"
            f"{quote(permission_name)}/",
            AssetPermissionCheckResponse,
            data={
                "artifact_id": artifact_id,
                "work_request_id": work_request_id,
                "workspace": workspace,
            },
        )

    @overload
    def lookup_single(
        self,
        lookup: LookupSingle,
        work_request: int,
        workspace: str,
        expect_type: Literal[LookupChildType.ARTIFACT],
        default_category: CollectionCategory | None = None,
    ) -> LookupSingleResponseArtifact: ...

    @overload
    def lookup_single(
        self,
        lookup: LookupSingle,
        work_request: int,
        workspace: str,
        expect_type: Literal[LookupChildType.COLLECTION],
        default_category: CollectionCategory | None = None,
    ) -> LookupSingleResponseCollection: ...

    @overload
    def lookup_single(
        self,
        lookup: LookupSingle,
        work_request: int,
        workspace: str,
        expect_type: LookupChildType,
        default_category: CollectionCategory | None = None,
    ) -> LookupSingleResponse: ...

    def lookup_single(
        self,
        lookup: LookupSingle,
        work_request: int,
        workspace: str,
        expect_type: LookupChildType,
        default_category: CollectionCategory | None = None,
    ) -> LookupSingleResponse:
        """Look up a single collection item."""
        request = LookupSingleRequest(
            lookup=lookup,
            work_request=work_request,
            workspace=workspace,
            expect_type=expect_type,
            default_category=default_category,
        )
        return self._debusine_http_client.post(
            "/lookup/single/", LookupSingleResponse, data=request.model_dump()
        )

    @overload
    def lookup_multiple(
        self,
        lookup: LookupMultiple,
        work_request: int,
        workspace: str,
        expect_type: Literal[LookupChildType.ARTIFACT],
        default_category: CollectionCategory | None = None,
    ) -> LookupMultipleResponse[LookupSingleResponseArtifact]: ...

    @overload
    def lookup_multiple(
        self,
        lookup: LookupMultiple,
        work_request: int,
        workspace: str,
        expect_type: Literal[LookupChildType.COLLECTION],
        default_category: CollectionCategory | None = None,
    ) -> LookupMultipleResponse[LookupSingleResponseCollection]: ...

    @overload
    def lookup_multiple(
        self,
        lookup: LookupMultiple,
        work_request: int,
        workspace: str,
        expect_type: LookupChildType,
        default_category: CollectionCategory | None = None,
    ) -> LookupMultipleResponse[LookupSingleResponse]: ...

    def lookup_multiple(
        self,
        lookup: LookupMultiple,
        work_request: int,
        workspace: str,
        expect_type: LookupChildType,
        default_category: CollectionCategory | None = None,
    ) -> LookupMultipleResponse[LookupSingleResponse]:
        """Look up multiple collection items."""
        request = LookupMultipleRequest(
            lookup=lookup,
            work_request=work_request,
            workspace=workspace,
            expect_type=expect_type,
            default_category=default_category,
        )
        return self._debusine_http_client.post(
            "/lookup/multiple/",
            LookupMultipleResponse,
            data=request.model_dump(),
        )

    def fetch_task_configuration_collection(
        self, *, workspace: str, name: str
    ) -> "RemoteTaskConfigurationRepository":
        """Fetch a task configuration collection."""
        from debusine.client.task_configuration import (
            Manifest,
            RemoteTaskConfigurationRepository,
        )

        response = self._debusine_http_client.get(
            f"/task-configuration/{quote(workspace)}/{quote(name)}/",
            TaskConfigurationCollectionContents,
        )
        manifest = Manifest(workspace=workspace, collection=response.collection)
        return RemoteTaskConfigurationRepository.from_items(
            manifest, response.items
        )

    def push_task_configuration_collection(
        self,
        *,
        repo: "LocalTaskConfigurationRepository",
        dry_run: bool,
    ) -> TaskConfigurationCollectionUpdateResults:
        """Replace the contents of a task configuration collection on server."""
        from debusine.client.task_configuration import InvalidRepository

        if repo.manifest is None:
            raise InvalidRepository("repository has no manifest")

        data = {
            "collection": repo.manifest.collection.model_dump(),
            "items": [i.model_dump() for i in repo.iter_entries()],
            "dry_run": dry_run,
        }

        return self._debusine_http_client.post(
            f"/task-configuration/{quote(repo.manifest.workspace)}"
            f"/{quote(repo.manifest.collection.name)}/",
            data=data,
            expected_class=TaskConfigurationCollectionUpdateResults,
        )

    def get_workspace_inheritance(
        self, workspace: str
    ) -> WorkspaceInheritanceChain:
        """Get a workspace inheritance chain."""
        return self._debusine_http_client.get(
            f"/workspace/{quote(workspace)}/inheritance/",
            expected_class=WorkspaceInheritanceChain,
        )

    def set_workspace_inheritance(
        self, workspace: str, *, chain: WorkspaceInheritanceChain
    ) -> WorkspaceInheritanceChain:
        """Set a workspace inheritance chain."""
        return self._debusine_http_client.post(
            f"/workspace/{quote(workspace)}/inheritance/",
            data=chain.model_dump(mode="json"),
            expected_class=WorkspaceInheritanceChain,
        )

    def collection_get(
        self, workspace: str, category: str, name: str, with_stats: bool = False
    ) -> CollectionData:
        """Fetch information about a collection."""
        url = f"/collection/{quote(workspace)}/{quote(category)}/{quote(name)}/"
        if with_stats:
            url += "?stats=1"
        return self._debusine_http_client.get(
            url, expected_class=CollectionData
        )

    def collection_create(
        self, workspace: str, data: CollectionDataNew
    ) -> CollectionData:
        """Create a new collection."""
        return self._debusine_http_client.post(
            f"/collection/{quote(workspace)}/",
            data=data.model_dump(mode="json"),
            expected_class=CollectionData,
            expected_statuses=[requests.codes.created],
        )

    def collection_update(
        self, workspace: str, data: CollectionData
    ) -> CollectionData:
        """Set information for a collection."""
        return self._debusine_http_client.patch(
            f"/collection/{quote(workspace)}/{data.id}/",
            data=data.model_dump(mode="json"),
            expected_class=CollectionData,
        )

    def collection_delete(self, workspace: str, data: CollectionData) -> None:
        """Delete a collection."""
        self._debusine_http_client.delete(
            f"/collection/{quote(workspace)}/{data.id}/",
            expected_statuses=[requests.codes.no_content],
        )

    def collection_iter(self, workspace: str) -> Iterable[CollectionData]:
        """List collections on a workspace."""
        return self._debusine_http_client.iter_paginated_get(
            f"/collection/{quote(workspace)}/",
            CollectionData,
        )

    def collection_token_grant_iter(
        self,
        *,
        workspace: str,
        collection_category: CollectionCategory,
        collection_name: str,
    ) -> Iterable[CollectionTokenGrantResponse]:
        """List tokens that have been granted roles on a collection."""
        return self._debusine_http_client.iter_paginated_get(
            f"/collection/{quote(workspace)}/{quote(collection_category)}"
            f"/{quote(collection_name)}/token/",
            CollectionTokenGrantResponse,
        )

    def collection_token_grant_create(
        self,
        *,
        workspace: str,
        collection_category: CollectionCategory,
        collection_name: str,
        request: CollectionTokenGrantRequest,
    ) -> CollectionTokenGrantResponse:
        """Grant a role on a collection to a token or an external username."""
        return self._debusine_http_client.post(
            f"/collection/{quote(workspace)}/{quote(collection_category)}"
            f"/{quote(collection_name)}/token/",
            CollectionTokenGrantResponse,
            data=request.model_dump(exclude_unset=True),
            expected_statuses=[requests.codes.created],
        )

    def collection_token_grant_delete(
        self,
        *,
        workspace: str,
        collection_category: CollectionCategory,
        collection_name: str,
        token_hash: str,
        role: CollectionRoles,
    ) -> None:
        """Revoke a role on a collection from a token."""
        self._debusine_http_client.delete(
            f"/collection/{quote(workspace)}/{quote(collection_category)}"
            f"/{quote(collection_name)}/token/{quote(token_hash)}"
            f"/{quote(role)}/",
            expected_statuses=[requests.codes.no_content],
        )

    def emit_metric(
        self,
        *,
        metric_type: Literal["counter", "gauge", "summary", "histogram"],
        name: str,
        labels: dict[str, str],
        value: float,
    ) -> None:
        """Ask the server to emit a metric on behalf of this client."""
        self._debusine_http_client.post(
            "/open-metrics/emit/",
            data=EmitMetricRequest(
                metric_type=metric_type, name=name, labels=labels, value=value
            ).model_dump(mode="json"),
            expected_class=None,
            expected_statuses=[requests.codes.no_content],
        )

    def server_notifications(self, endpoint: str) -> "ServerNotifications":
        """Open a server notifications stream from the given endpoint."""
        from debusine.client.server_notifications import ServerNotifications

        return ServerNotifications(self, endpoint=endpoint)

    def _on_work_request_completed(
        self,
        *,
        command: Sequence[str | os.PathLike[str]],
        working_directory: Path,
        last_completed_at: Path | None,
        msg_content: dict[str, Any],
    ) -> None:
        on_work_request_completed = OnWorkRequestCompleted.model_validate(
            msg_content
        )

        # Execute the command
        cmd = [
            *command,
            str(on_work_request_completed.work_request_id),
            on_work_request_completed.result,
        ]
        self._logger.info("Executing %s", cmd)
        # TODO: this blocks the asyncio main loop: if the process takes longer
        # than a minute, the websocket client will likely disconnect. It can be
        # replaced with asyncio.subprocess
        p = subprocess.Popen(cmd, cwd=working_directory)
        p.wait()

        if last_completed_at is not None:
            self.write_last_completed_at(
                last_completed_at,
                on_work_request_completed.completed_at,
            )

    @staticmethod
    def write_last_completed_at(
        completed_at_file: Path,
        completed_at: dt.datetime | None,
    ) -> None:
        """Write to completed_at_file the completed_at datetime."""
        completed_at_formatted = (
            completed_at.isoformat() if completed_at else None
        )

        completed_at_file.write_text(
            json.dumps({"last_completed_at": completed_at_formatted}, indent=2)
            + "\n"
        )

    @staticmethod
    def _read_last_completed_at(
        last_completed_at: Path | None,
    ) -> str | None:
        if last_completed_at is None:
            return None

        if not last_completed_at.exists():
            # The file does not exist, it will be created if a WorkRequest
            # is completed
            return None

        last_completed_at_json = json.loads(last_completed_at.read_bytes())
        completed_at = last_completed_at_json.get("last_completed_at")
        assert completed_at is None or isinstance(completed_at, str)
        return completed_at

    def on_work_request_completed(
        self,
        *,
        workspaces: list[str] | None = None,
        last_completed_at: Path | None = None,
        command: list[str],
        working_directory: Path,
    ) -> None:
        """Execute command when a work request is completed."""
        asyncio.run(
            self.on_work_request_completed_async(
                workspaces=workspaces,
                last_completed_at=last_completed_at,
                command=command,
                working_directory=working_directory,
            )
        )

    async def on_work_request_completed_async(
        self,
        *,
        workspaces: list[str] | None = None,
        last_completed_at: Path | None = None,
        command: list[str],
        working_directory: Path,
    ) -> None:
        """Execute command when a work request is completed."""
        # Prevent import loop
        params = {}

        if workspaces:
            params["workspaces"] = ",".join(workspaces)

        completed_at_since = self._read_last_completed_at(last_completed_at)
        if completed_at_since:
            params["completed_at_since"] = completed_at_since

        endpoint = "1.0/work-request/on-completed/"
        if params:
            endpoint += "?" + urlencode(params)

        async with self.server_notifications(
            endpoint=endpoint
        ) as server_notifications:
            async for payload in server_notifications.messages():
                match payload.get("text"):
                    case "connected":
                        self._logger.info("Connected!")
                    case "work_request_completed":
                        self._on_work_request_completed(
                            command=command,
                            working_directory=working_directory,
                            last_completed_at=last_completed_at,
                            msg_content=payload,
                        )
                    case _:
                        self._logger.warning(
                            "invalid text in msg_content"
                            " received from the server: %r ",
                            payload.get("text"),
                        )

    def decode_ws_message(
        self, msg: aiohttp.http_websocket.WSMessage
    ) -> dict[str, Any]:
        """
        Decode a websocket message received from the server.

        If ``msg.type`` is aiohttp.WSMsgType.TEXT: decode msg.data (contains
        JSON) and return it.

        If ``msg`` contains an error condition, raise it or log it
        appropriately.

        :returns: the decoded payload, or None if the message contained an
          error that got logged.

        :raise ServerNotificationDecodeError: for malformed messages
        :raise TokenDisabledError: if reason_code is TOKEN_DISABLED.
        """
        if not isinstance(msg, aiohttp.http_websocket.WSMessage):
            raise TypeError(
                f"Message {msg!r} is not an instance of"
                " aiohttp.http_websocket.WSMessage",
            )

        if msg.type != aiohttp.WSMsgType.TEXT:
            raise ServerNotificationDecodeError("Message is not a TEXT message")

        try:
            msg_content = json.loads(msg.data)
        except json.JSONDecodeError as exc:
            raise ServerNotificationDecodeError(
                f"Message {msg.data!r} does not decode as JSON: {exc}"
            )

        if not isinstance(msg_content, dict):
            raise ServerNotificationDecodeError(
                f"JSON message {msg_content!r} does not contain a dict"
            )

        if msg_content.get('text') is not None:
            # Ordinary message
            return msg_content
        elif (reason_code := msg_content.get('reason_code')) is not None:
            # Error message
            if reason_code == 'TOKEN_DISABLED':
                raise TokenDisabledError(msg_content.get("reason", "Unknown"))
            else:
                raise ServerNotificationDecodeError(
                    f"Unsupported error reason code {reason_code!r}"
                )
        else:
            raise ServerNotificationDecodeError(
                f"Unsupported message payload {msg_content!r}"
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command group."""

from typing import ClassVar, override
from unittest import mock

from django.core.management import CommandError

from debusine.db.models import Group, Scope, User
from debusine.db.models.auth import GroupRoles
from debusine.db.playground import scenarios
from debusine.django.management.tests import call_command
from debusine.server.management.commands.group import (
    Command,
    ParsedGroupArgument,
)
from debusine.server.management.commands.tests.utils import TabularOutputTests
from debusine.test.django import TestCase


class ParsedGroupArgumentTests(TestCase):
    """Tests for :py:class:`ParsedGroupArgument` class."""

    scenario = scenarios.DefaultContext()
    group: ClassVar[Group]
    workspace_group: ClassVar[Group]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.group = cls.playground.create_group("group")
        cls.workspace_group = cls.playground.create_group(
            "group", workspace=cls.scenario.workspace
        )

    def test_str(self) -> None:
        for value in ("foo/bar", "foo/bar/baz"):
            with self.subTest(value=value):
                parsed = ParsedGroupArgument.parse(value)
                self.assertEqual(str(parsed), value)
                self.assertEqual(repr(parsed), f"ParsedGroupArgument({value})")

    def test_parse(self) -> None:
        for value, scope, workspace, name in (
            ("foo/bar", "foo", None, "bar"),
            ("foo/bar/baz", "foo", "bar", "baz"),
        ):
            with self.subTest(value=value):
                parsed = ParsedGroupArgument.parse(value)
                self.assertEqual(parsed.scope, scope)
                self.assertEqual(parsed.workspace, workspace)
                self.assertEqual(parsed.name, name)

    def test_as_queryset_filter(self) -> None:
        for value, expected in (
            (
                "foo/bar",
                {"scope__name": "foo", "workspace": None, "name": "bar"},
            ),
            (
                "foo/bar/baz",
                {"scope__name": "foo", "workspace__name": "bar", "name": "baz"},
            ),
        ):
            with self.subTest(value=value):
                parsed = ParsedGroupArgument.parse(value)
                self.assertEqual(parsed.as_queryset_filter(), expected)

    def test_scope_instance(self) -> None:
        parsed = ParsedGroupArgument.parse("debusine/name")
        self.assertEqual(parsed.scope_instance, self.scenario.scope)

    def test_scope_instance_invalid(self) -> None:
        parsed = ParsedGroupArgument.parse("invalid/name")
        with self.assertRaisesRegex(
            CommandError, "Scope 'invalid' not found"
        ) as exc:
            parsed.scope_instance
        self.assertEqual(getattr(exc.exception, "returncode"), 3)

    def test_workspace_instance(self) -> None:
        parsed = ParsedGroupArgument.parse("debusine/System/name")
        self.assertEqual(parsed.workspace_instance, self.scenario.workspace)

    def test_workspace_instance_unset(self) -> None:
        parsed = ParsedGroupArgument.parse("debusine/name")
        self.assertIsNone(parsed.workspace_instance)

    def test_workspace_instance_invalid(self) -> None:
        parsed = ParsedGroupArgument.parse("debusine/invalid/name")
        with self.assertRaisesRegex(
            CommandError, "Workspace 'debusine/invalid' not found"
        ) as exc:
            parsed.workspace_instance
        self.assertEqual(getattr(exc.exception, "returncode"), 3)

    def test_group_instance(self) -> None:
        parsed = ParsedGroupArgument.parse("debusine/group")
        self.assertEqual(parsed.group_instance, self.group)

    def test_group_instance_workspaced(self) -> None:
        parsed = ParsedGroupArgument.parse("debusine/System/group")
        self.assertEqual(parsed.group_instance, self.workspace_group)

    def test_group_instance_invalid(self) -> None:
        parsed = ParsedGroupArgument.parse("debusine/System/invalid")
        with self.assertRaisesRegex(
            CommandError,
            "Group 'invalid' not found in scope 'debusine'"
            " and workspace 'System'",
        ) as exc:
            parsed.group_instance
        self.assertEqual(getattr(exc.exception, "returncode"), 3)


class GroupCommandTests(TabularOutputTests, TestCase):
    """Tests for group create management command."""

    scenario = scenarios.DefaultContext()
    group: ClassVar[Group]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.group = cls.playground.create_group("group")

    def assertGroup(
        self, scope: Scope, name: str, users: list[User] | None = None
    ) -> Group:
        """Check that the given scope exists, with its Admin group."""
        if users is None:
            users = []
        group = Group.objects.get(scope=scope, name=name)
        self.assertQuerySetEqual(group.users.all(), users, ordered=False)
        return group

    def test_get_scope_and_workspace(self) -> None:
        command = Command()
        for arg, scope, workspace in (
            ("debusine", self.scenario.scope, None),
            (
                "debusine/System",
                self.scenario.scope,
                self.scenario.workspace,
            ),
        ):
            with self.subTest(arg=arg):
                arg_scope, arg_workspace = command.get_scope_and_workspace(arg)
                self.assertEqual(arg_scope, scope)
                self.assertEqual(arg_workspace, workspace)

    def test_invalid_action(self) -> None:
        """Test invoking an invalid subcommand."""
        with self.assertRaisesRegex(
            CommandError, r"invalid choice: 'does-not-exist'"
        ) as exc:
            call_command("group", "does-not-exist")

        self.assertEqual(getattr(exc.exception, "returncode"), 1)

    def test_unexpected_action(self) -> None:
        """Test a subcommand with no implementation."""
        command = Command()

        with self.assertRaisesRegex(
            CommandError, r"Action 'does_not_exist' not found"
        ) as exc:
            command.handle(action="does_not_exist")

        self.assertEqual(getattr(exc.exception, "returncode"), 3)

    def test_get_users(self) -> None:
        """Test get_users."""
        user1 = self.playground.create_user("user1")
        user2 = self.playground.create_user("user2")

        command = Command()
        self.assertEqual(command.get_users(["user1", "user2"]), [user1, user2])

    def test_get_users_empty(self) -> None:
        """Test get_users with an empty username list."""
        command = Command()
        self.assertEqual(command.get_users([]), [])

    def test_get_users_nonexisting(self) -> None:
        """Test get_users with nonexisting usernames."""
        command = Command()
        with self.assertRaisesRegex(
            CommandError,
            r"User 'missing1' does not exist\n"
            r"User 'missing2' does not exist",
        ) as exc:
            command.get_users(["missing1", "missing2"])
        self.assertEqual(getattr(exc.exception, "returncode"), 3)

    def test_list_no_users(self) -> None:
        with self.assertPrintsTable() as output:
            _, stderr, exit_code = call_command("group", "list", "debusine")

        self.assertEqual(output.col(0), ["debusine"])
        self.assertEqual(output.col(1), ["-"])
        self.assertEqual(output.col(2), ["group"])
        self.assertEqual(output.col(3), ["0"])

        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

    def test_list(self) -> None:
        self.playground.create_group("group", workspace=self.scenario.workspace)
        self.playground.add_user(self.group, self.scenario.user)
        with self.assertPrintsTable() as output:
            _, stderr, exit_code = call_command("group", "list", "debusine")

        self.assertEqual(output.col(0), ["debusine", "debusine"])
        self.assertEqual(output.col(1), ["-", "System"])
        self.assertEqual(output.col(2), ["group", "group"])
        self.assertEqual(output.col(3), ["1", "0"])

        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

    def test_list_workspace(self) -> None:
        self.playground.create_group("group", workspace=self.scenario.workspace)
        with self.assertPrintsTable() as output:
            _, stderr, exit_code = call_command(
                "group", "list", "debusine/System"
            )

        self.assertEqual(output.col(0), ["debusine"])
        self.assertEqual(output.col(1), ["System"])
        self.assertEqual(output.col(2), ["group"])
        self.assertEqual(output.col(3), ["0"])

        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

    def test_list_no_workspace(self) -> None:
        self.playground.add_user(self.group, self.scenario.user)
        self.playground.create_group("group", workspace=self.scenario.workspace)
        with self.assertPrintsTable() as output:
            _, stderr, exit_code = call_command(
                "group", "list", "debusine", "--no-workspace"
            )

        self.assertEqual(output.col(0), ["debusine"])
        self.assertEqual(output.col(1), ["-"])
        self.assertEqual(output.col(2), ["group"])
        self.assertEqual(output.col(3), ["1"])

        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

    def test_create(self) -> None:
        """Test a successful create."""
        stdout, stderr, exit_code = call_command(
            "group", "create", "debusine/group1"
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        self.assertGroup(self.scenario.scope, "group1")

    def test_create_idempotent(self) -> None:
        """Test a idempotence in create twice."""
        stdout, stderr, exit_code = call_command(
            "group", "create", "debusine/group1"
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        group = self.assertGroup(self.scenario.scope, "group1")
        self.playground.add_user(group, self.playground.get_default_user())

        stdout, stderr, exit_code = call_command(
            "group", "create", "debusine/group1"
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        group1 = self.assertGroup(
            self.scenario.scope, "group1", [self.playground.get_default_user()]
        )

        self.assertEqual(group.pk, group1.pk)

    def test_create_invalid_argument(self) -> None:
        with self.assertRaisesRegex(
            CommandError,
            r"should be in the form 'scopename\[/workspacename\]/groupname",
        ) as exc:
            call_command("group", "create", "debusine/workspace/group/invalid")
        self.assertEqual(getattr(exc.exception, "returncode"), 3)
        self.assertQuerySetEqual(Group.objects.filter(name="group/invalid"), [])

    def test_create_invalid_name(self) -> None:
        stdout, stderr, exit_code = call_command(
            "group", "create", "debusine/@name"
        )
        self.assertEqual(stdout, "")
        self.assertEqual(
            stderr,
            "Created group would be invalid:\n"
            "* name: '@name' is not a valid group name\n",
        )
        self.assertEqual(exit_code, 3)
        self.assertQuerySetEqual(Group.objects.filter(name="@name"), [])

    def test_rename(self) -> None:
        """Test a successful rename."""
        self.playground.add_user(self.group, self.playground.get_default_user())
        stdout, stderr, exit_code = call_command(
            "group", "rename", "debusine/group", "group1"
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        self.assertGroup(
            self.scenario.scope, "group1", [self.playground.get_default_user()]
        )

    def test_rename_noop(self) -> None:
        """Test renaming to current name."""
        self.playground.add_user(self.group, self.playground.get_default_user())

        with mock.patch("debusine.db.models.Group.save") as save:
            stdout, stderr, exit_code = call_command(
                "group", "rename", "debusine/group", "group"
            )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        self.assertGroup(
            self.scenario.scope, "group", [self.playground.get_default_user()]
        )

        save.assert_not_called()

    def test_rename_group_does_not_exist(self) -> None:
        """Test renaming a nonexisting group."""
        with self.assertRaisesRegex(
            CommandError, r"Group 'does-not-exist' not found"
        ) as exc:
            call_command(
                "group", "rename", "debusine/does-not-exist", "newname"
            )

        self.assertEqual(getattr(exc.exception, "returncode"), 3)

    def test_rename_target_exists(self) -> None:
        """Test renaming with a name already in use."""
        group = self.playground.create_group("test")
        stdout, stderr, exit_code = call_command(
            "group", "rename", "debusine/test", "group"
        )
        self.assertEqual(stdout, "")
        self.assertEqual(
            stderr.splitlines(),
            [
                "Renamed group would be invalid:",
                (
                    "* Constraint “db_group_unique_name_scope_workspace”"
                    " is violated."
                ),
                # From django 5.0, when db_group_unique_name_scope_workspace is
                # redone with nulls_distinct, we can go back to a better error
                # message:
                # "* Group with this Name and Scope already exists.",
            ],
        )

        self.assertEqual(exit_code, 3)
        group.refresh_from_db()
        self.assertEqual(group.name, "test")

    def test_rename_new_name_invalid(self) -> None:
        """Test renaming to an invalid name."""
        stdout, stderr, exit_code = call_command(
            "group", "rename", "debusine/group", "invalid/name"
        )
        self.assertEqual(stdout, "")
        self.assertEqual(
            stderr.splitlines(),
            [
                "Renamed group would be invalid:",
                "* name: 'invalid/name' is not a valid group name",
            ],
        )
        self.assertEqual(exit_code, 3)
        self.group.refresh_from_db()
        self.assertEqual(self.group.name, "group")

    def test_delete(self) -> None:
        """Test a successful delete."""
        self.playground.add_user(self.group, self.playground.get_default_user())
        stdout, stderr, exit_code = call_command(
            "group",
            "delete",
            "debusine/group",
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        self.assertQuerySetEqual(
            Group.objects.filter(scope=self.scenario.scope), []
        )

    def test_delete_missing(self) -> None:
        """Test idempotence in deleting a missing group."""
        self.playground.add_user(self.group, self.playground.get_default_user())
        stdout, stderr, exit_code = call_command(
            "group",
            "delete",
            "debusine/missing",
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        self.assertQuerySetEqual(
            Group.objects.filter(scope=self.scenario.scope), [self.group]
        )

    def test_members_set(self) -> None:
        """Test members --set."""
        user1 = self.playground.create_user("user1")
        user2 = self.playground.create_user("user2")
        self.playground.add_user(self.group, self.playground.get_default_user())

        stdout, stderr, exit_code = call_command(
            "group", "members", "debusine/group", "--set", "user1", "user2"
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        self.assertQuerySetEqual(
            self.group.users.all(), [user1, user2], ordered=False
        )

    def test_members_set_empty(self) -> None:
        """Test members --set to the empty set."""
        self.playground.add_user(self.group, self.playground.get_default_user())

        stdout, stderr, exit_code = call_command(
            "group", "members", "debusine/group", "--set"
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        self.assertQuerySetEqual(self.group.users.all(), [])

    def test_members_add(self) -> None:
        """Test members --add."""
        user1 = self.playground.create_user("user1")
        self.playground.add_user(self.group, self.playground.get_default_user())

        stdout, stderr, exit_code = call_command(
            "group",
            "members",
            "debusine/group",
            "--add",
            "user1",
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        self.assertQuerySetEqual(
            self.group.users.all(),
            [self.playground.get_default_user(), user1],
            ordered=False,
        )

    def test_members_remove(self) -> None:
        """Test members --remove."""
        user1 = self.playground.create_user("user1")
        self.playground.add_user(self.group, self.playground.get_default_user())
        self.playground.add_user(self.group, user1)

        stdout, stderr, exit_code = call_command(
            "group",
            "members",
            "debusine/group",
            "--remove",
            self.playground.get_default_user().username,
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        self.assertQuerySetEqual(
            self.group.users.all(),
            [user1],
        )

    def test_members_mix(self) -> None:
        """Test running multiple members actions."""
        with self.assertRaisesRegex(
            CommandError,
            r"Error: argument --add: not allowed with argument --remove",
        ) as exc:
            call_command(
                "group",
                "members",
                "debusine/group",
                "--remove",
                "1",
                "--add",
                "2",
            )
        self.assertEqual(getattr(exc.exception, "returncode"), 1)

    def test_members_list_empty(self) -> None:
        """Test group members with no users."""
        with self.assertPrintsTable() as output:
            _, stderr, exit_code = call_command(
                "group", "members", "debusine/group"
            )

        self.assertEqual(output.col(0), [])
        self.assertEqual(output.col(1), [])
        self.assertEqual(output.col(2), [])
        self.assertEqual(output.col(3), [])

        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

    def test_members_list(self) -> None:
        """Test group members."""
        self.playground.add_user(self.group, self.scenario.user)
        with self.assertPrintsTable() as output:
            _, stderr, exit_code = call_command(
                "group", "members", "debusine/group"
            )

        self.assertEqual(output.col(0), [self.scenario.user.username])
        self.assertEqual(output.col(1), [self.scenario.user.email])
        self.assertEqual(output.col(2), [GroupRoles.MEMBER])
        self.assertEqual(
            output.col(3), [self.scenario.user.date_joined.isoformat()]
        )

        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

    def test_members_list_multiple_roles(self) -> None:
        """Test group members list shows correct roles for each member."""
        user1 = self.playground.create_user("user1")
        user2 = self.playground.create_user("user2")
        self.playground.add_user(self.group, user1, role=GroupRoles.MEMBER)
        self.playground.add_user(self.group, user2, role=GroupRoles.ADMIN)

        with self.assertPrintsTable() as output:
            _, stderr, exit_code = call_command(
                "group", "members", "debusine/group"
            )

        # Rows ordered by user__date_joined
        self.assertEqual(output.col(0), [user1.username, user2.username])
        self.assertEqual(output.col(2), [GroupRoles.MEMBER, GroupRoles.ADMIN])

        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

    def test_members_add_default_role(self) -> None:
        """Test --add without --role. default is MEMBER."""
        user1 = self.playground.create_user("user1")

        stdout, stderr, exit_code = call_command(
            "group", "members", "debusine/group", "--add", "user1"
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        self.assertEqual(self.group.get_user_role(user1), GroupRoles.MEMBER)

    def test_members_add_member_role_explicit(self) -> None:
        """Test --add --role member explicitly sets MEMBER role"""
        user1 = self.playground.create_user("user1")

        stdout, stderr, exit_code = call_command(
            "group",
            "members",
            "debusine/group",
            "--add",
            "user1",
            "--role",
            "member",
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        self.assertEqual(self.group.get_user_role(user1), GroupRoles.MEMBER)

    def test_members_add_admin_role(self) -> None:
        """Test --add --role admin creates a membership with ADMIN role"""
        user1 = self.playground.create_user("user1")

        stdout, stderr, exit_code = call_command(
            "group",
            "members",
            "debusine/group",
            "--add",
            "user1",
            "--role",
            "admin",
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        self.assertEqual(self.group.get_user_role(user1), GroupRoles.ADMIN)

    def test_members_add_multiple_users_same_role(self) -> None:
        """Test --add with multiple users assigns the same role to all"""
        user1 = self.playground.create_user("user1")
        user2 = self.playground.create_user("user2")

        _, _, exit_code = call_command(
            "group",
            "members",
            "debusine/group",
            "--add",
            "user1",
            "user2",
            "--role",
            "admin",
        )
        self.assertEqual(exit_code, 0)

        self.assertEqual(self.group.get_user_role(user1), GroupRoles.ADMIN)
        self.assertEqual(self.group.get_user_role(user2), GroupRoles.ADMIN)

    def test_members_set_admin_role(self) -> None:
        """Test --set --role admin replaces all members with ADMIN role"""
        user1 = self.playground.create_user("user1")
        user2 = self.playground.create_user("user2")
        self.playground.add_user(self.group, self.playground.get_default_user())

        stdout, stderr, exit_code = call_command(
            "group",
            "members",
            "debusine/group",
            "--set",
            "user1",
            "user2",
            "--role",
            "admin",
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        self.assertEqual(self.group.get_user_role(user1), GroupRoles.ADMIN)
        self.assertEqual(self.group.get_user_role(user2), GroupRoles.ADMIN)
        # Previous member should be gone
        self.assertNotIn(
            self.playground.get_default_user(), self.group.users.all()
        )

    def test_members_role_invalid(self) -> None:
        """Test invalid --role value is rejected"""
        with self.assertRaisesRegex(
            CommandError,
            r"invalid choice: 'superuser'",
        ) as exc:
            call_command(
                "group",
                "members",
                "debusine/group",
                "--add",
                "someuser",
                "--role",
                "superuser",
            )
        self.assertEqual(getattr(exc.exception, "returncode"), 1)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""URLs related to workspaces."""

from django.urls import URLPattern, URLResolver, include, path

import debusine.web.views.workspaces as views

app_name = "workspaces"

workspace_urlpatterns: list[URLPattern | URLResolver] = [
    path(
        "",
        views.WorkspaceDetailView.as_view(),
        name="detail",
    ),
    path(
        "configure/",
        views.WorkspaceUpdateView.as_view(),
        name="update",
    ),
    path(
        "inheritance/",
        views.WorkspaceUpdateInheritanceView.as_view(),
        name="update_inheritance",
    ),
    path(
        "collection/",
        include(
            "debusine.web.urls.collections",
            namespace="collections",
        ),
    ),
    path(
        "workflow/",
        include(
            "debusine.web.urls.workflows",
            namespace="workflows",
        ),
    ),
    path(
        "workflow-template/",
        include(
            "debusine.web.urls.workflow_templates",
            namespace="workflow_templates",
        ),
    ),
    path(
        "artifact/",
        include(
            "debusine.web.urls.artifacts",
            namespace="artifacts",
        ),
    ),
    path(
        "work-request/",
        include("debusine.web.urls.work_requests", namespace="work-requests"),
    ),
    path(
        "task-configuration-inspector/",
        views.TaskConfigInspectorView.as_view(),
        name="task_configuration_inspector",
    ),
    path(
        "groups/",
        include("debusine.web.urls.groups", namespace="groups"),
    ),
]

urlpatterns = [
    path(
        "<str:wname>/",
        include(workspace_urlpatterns),
    ),
]
